"""Classes and objects for drawing elements (SVG-like)"""

import abc
import dataclasses
import math
import copy
import enum
import typing
import typing_extensions
import collections.abc

import shapely.geometry
import shapely.affinity
import shapely.ops

import momapy.geometry
import momapy.coloring
import momapy.utils


class NoneValueType(object):  # TODO: implement true singleton
    """Class for none values (as in SVG)"""

    def __copy__(self):
        return self

    def __deepcopy__(self, memo):
        return self

    def __eq__(self, other):
        return type(self) is type(other)

    def __hash__(self):
        return id(NoneValue)


NoneValue = NoneValueType()
"""A singleton value for type `NoneValueType`"""


@dataclasses.dataclass(frozen=True)
class FilterEffect(abc.ABC):
    """Class for filter effects"""

    result: str | None = dataclasses.field(
        default=None, metadata={"description": "The name of the result"}
    )


@dataclasses.dataclass(frozen=True, kw_only=True)
class DropShadowEffect(FilterEffect):
    """Class for drop shadow effects"""

    dx: float = dataclasses.field(
        default=0.0,
        metadata={"description": "The horizontal offset of the shadow"},
    )
    dy: float = dataclasses.field(
        default=0.0,
        metadata={"description": "The vertical offset of the shadow"},
    )
    std_deviation: float = dataclasses.field(
        default=0.0,
        metadata={
            "description": "The standard deviation to be used to compute the shadow"
        },
    )
    flood_opacity: float = dataclasses.field(
        default=1.0,
        metadata={"description": "The flood opacity of the shadow"},
    )
    flood_color: momapy.coloring.Color = dataclasses.field(
        default=momapy.coloring.black,
        metadata={"description": "The color of the shadow"},
    )

    def to_compat(self) -> list[FilterEffect]:
        """Return a list of filter effects better supported by software that is equivalent to the given drop-shadow effect"""
        flood_effect = FloodEffect(
            result=momapy.utils.make_uuid4_as_str(),
            flood_opacity=self.flood_opacity,
            flood_color=self.flood_color,
        )
        composite_effect1 = CompositeEffect(
            in_=flood_effect.result,
            in2=FilterEffectInput.SOURCE_GRAPHIC,
            operator=CompositionOperator.IN,
            result=momapy.utils.make_uuid4_as_str(),
        )
        gaussian_blur_effect = GaussianBlurEffect(
            in_=composite_effect1.result,
            std_deviation=self.std_deviation,
            result=momapy.utils.make_uuid4_as_str(),
        )
        offset_effect = OffsetEffect(
            in_=gaussian_blur_effect.result,
            dx=self.dx,
            dy=self.dy,
            result=momapy.utils.make_uuid4_as_str(),
        )
        composite_effect2 = CompositeEffect(
            in_=FilterEffectInput.SOURCE_GRAPHIC,
            in2=offset_effect.result,
            operator=CompositionOperator.OVER,
            result=self.result,
        )
        effects = [
            flood_effect,
            composite_effect1,
            gaussian_blur_effect,
            offset_effect,
            composite_effect2,
        ]
        return effects


class FilterEffectInput(enum.Enum):
    """Class for filter effect inputs"""

    SOURCE_GRAPHIC = 0
    SOURCE_ALPHA = 1
    BACKGROUND_IMAGE = 2
    BACKGROUND_ALPHA = 3
    FILL_PAINT = 4
    STROKE_PAINT = 5


class CompositionOperator(enum.Enum):
    """Class for composition operators"""

    OVER = 0
    IN = 1
    OUT = 2
    ATOP = 3
    XOR = 4
    LIGHTER = 5
    ARTIHMETIC = 6


@dataclasses.dataclass(frozen=True, kw_only=True)
class CompositeEffect(FilterEffect):
    """Class for composite effects"""

    in_: FilterEffectInput | str | None = dataclasses.field(
        default=None,
        metadata={
            "description": "The first effect input or the name of the first filter effect input"
        },
    )
    in2: FilterEffectInput | str | None = dataclasses.field(
        default=None,
        metadata={
            "description": "The second filter effect input or the name of the second filter effect input"
        },
    )
    operator: CompositionOperator | None = dataclasses.field(
        default=CompositionOperator.OVER,
        metadata={
            "description": "The operator to be used to compute the composite effect"
        },
    )


@dataclasses.dataclass(frozen=True, kw_only=True)
class FloodEffect(FilterEffect):
    """Class for flood effects"""

    flood_color: momapy.coloring.Color = dataclasses.field(
        default=momapy.coloring.black,
        metadata={"description": "The color of the flood effect"},
    )
    flood_opacity: float = dataclasses.field(
        default=1.0,
        metadata={"description": "The opacity of the flood effect"},
    )


class EdgeMode(enum.Enum):
    """Class for edge modes"""

    DUPLICATE = 0
    WRAP = 1


@dataclasses.dataclass(frozen=True, kw_only=True)
class GaussianBlurEffect(FilterEffect):
    """Class for Gaussian blur effects"""

    in_: FilterEffectInput | str | None = None
    std_deviation: float = 0.0
    edge_mode: NoneValueType | EdgeMode = NoneValue


@dataclasses.dataclass(frozen=True, kw_only=True)
class OffsetEffect(FilterEffect):
    """Class for offset effects"""

    in_: FilterEffectInput | str | None = None
    dx: float = 0.0
    dy: float = 0.0


class FilterUnits(enum.Enum):
    """Class for filter units"""

    USER_SPACE_ON_USE = 0
    OBJECT_BOUNDING_BOX = 1


@dataclasses.dataclass(frozen=True, kw_only=True)
class Filter(object):
    """Class for filters"""

    id_: str = dataclasses.field(
        hash=False,
        compare=False,
        default_factory=momapy.utils.make_uuid4_as_str,
    )
    filter_units: FilterUnits = FilterUnits.OBJECT_BOUNDING_BOX
    effects: tuple[FilterEffect] = dataclasses.field(default_factory=tuple)
    width: float | str = "120%"
    height: float | str = "120%"
    x: float | str = "-10%"
    y: float | str = "-10%"

    def to_compat(self) -> typing_extensions.Self:
        """Return a filter equivalent to the given filter with its filter effects replaced by better supported"""
        effects = []
        for effect in self.effects:
            if hasattr(effect, "to_compat"):
                effects += effect.to_compat()
            else:
                effects.append(effect)
        return dataclasses.replace(self, effects=tuple(effects))


class FontStyle(enum.Enum):
    """Class for font styles"""

    NORMAL = 0
    ITALIC = 1
    OBLIQUE = 2


class FontWeight(enum.Enum):
    """Class for font weights"""

    NORMAL = 0
    BOLD = 1
    BOLDER = 2
    LIGHTER = 3


class TextAnchor(enum.Enum):
    """Class for text anchors"""

    START = 0
    MIDDLE = 1
    END = 2


# not implemented yet, just present in sbml render; to be implemented later
class FillRule(enum.Enum):
    """Class for fill rules"""

    NONZERO = 0
    EVENODD = 1


PRESENTATION_ATTRIBUTES = {
    "fill": {
        "initial": momapy.coloring.black,
        "inherited": True,
    },
    "fill_rule": {
        "initial": FillRule.NONZERO,
        "inherited": True,
    },
    "filter": {
        "initial": NoneValue,
        "inherited": False,
    },
    "font_family": {
        "initial": None,  # depends on the user agent
        "inherited": True,
    },
    "font_size": {
        "initial": None,  # medium, which depends on the user agent; usually 16px
        "inherited": True,
    },
    "font_style": {
        "initial": FontStyle.NORMAL,
        "inherited": True,
    },
    "font_weight": {
        "initial": FontWeight.NORMAL,
        "inherited": True,
    },
    "stroke": {
        "initial": NoneValue,
        "inherited": True,
    },
    "stroke_dasharray": {
        "initial": NoneValue,
        "inherited": True,
    },
    "stroke_dashoffset": {
        "initial": 0.0,
        "inherited": True,
    },
    "stroke_width": {
        "initial": 1.0,
        "inherited": True,
    },
    "text_anchor": {
        "initial": TextAnchor.START,
        "inherited": True,
    },
    "transform": {
        "initial": NoneValue,
        "inherited": False,
    },
}

INITIAL_VALUES = {
    "font_family": "Arial",
    "font_size": 16.0,
}

FONT_WEIGHT_TO_VALUE = {
    FontWeight.NORMAL: 400,
    FontWeight.BOLD: 700,
}


def get_initial_value(attr_name: str):
    """Return the initial value of a presentation attribute"""
    attr_value = PRESENTATION_ATTRIBUTES[attr_name]["initial"]
    if attr_value is None:
        attr_value = INITIAL_VALUES[attr_name]
    return attr_value


@dataclasses.dataclass(frozen=True, kw_only=True)
class DrawingElement(abc.ABC):
    """Class for drawing elements"""

    class_: str | None = dataclasses.field(
        default=None,
        metadata={"description": "The class name of the drawing element"},
    )
    fill: NoneValueType | momapy.coloring.Color | None = dataclasses.field(
        default=None,
        metadata={"description": "The fill color of the drawing element"},
    )
    fill_rule: FillRule | None = dataclasses.field(
        default=None,
        metadata={"description": "The fill rule of the drawing element"},
    )
    filter: NoneValueType | Filter | None = dataclasses.field(
        default=None,
        metadata={"description": "The filter of the drawing element"},
    )  # should be a tuple of filters to follow SVG (to be implemented)
    font_family: str | None = dataclasses.field(
        default=None,
        metadata={"description": "The font family of the drawing element"},
    )
    font_size: float | None = dataclasses.field(
        default=None,
        metadata={"description": "The font size of the drawing element"},
    )
    font_style: FontStyle | None = dataclasses.field(
        default=None,
        metadata={"description": "The font style of the drawing element"},
    )
    font_weight: FontWeight | int | None = dataclasses.field(
        default=None,
        metadata={"description": "The font weight of the drawing element"},
    )
    id_: str | None = dataclasses.field(
        default=None, metadata={"description": "The id of the drawing element"}
    )
    stroke: NoneValueType | momapy.coloring.Color | None = dataclasses.field(
        default=None,
        metadata={"description": "The stroke color of the drawing element"},
    )
    stroke_dasharray: NoneValueType | tuple[float, ...] | None = dataclasses.field(
        default=None,
        metadata={"description": "The stroke dasharray of the drawing element"},
    )
    stroke_dashoffset: NoneValueType | float | None = dataclasses.field(
        default=None,
        metadata={"description": "The stroke dashoffset of the drawing element"},
    )
    stroke_width: NoneValueType | float | None = dataclasses.field(
        default=None,
        metadata={"description": "The stroke width of the drawing element"},
    )
    text_anchor: TextAnchor | None = dataclasses.field(
        default=None,
        metadata={"description": "The text anchor of the drawing element"},
    )
    transform: NoneValueType | tuple[momapy.geometry.Transformation] | None = (
        dataclasses.field(
            default=None,
            metadata={"description": "The transform of the drawing element"},
        )
    )

    @abc.abstractmethod
    def to_shapely(self, to_polygons=False) -> shapely.GeometryCollection:
        """Compute and return a geometry collection built from the given drawing elements"""
        pass

    def bbox(self) -> momapy.geometry.Bbox:
        """Compute and return the bounding box of the drawing element"""
        bounds = self.to_shapely().bounds
        return momapy.geometry.Bbox(
            momapy.geometry.Point(
                (bounds[0] + bounds[2]) / 2, (bounds[1] + bounds[3]) / 2
            ),
            bounds[2] - bounds[0],
            bounds[3] - bounds[1],
        )

    def get_filter_region(self) -> momapy.geometry.Bbox:
        """Compute and return the filter region of the drawing element"""

        if self.filter is None or self.filter is NoneValue:
            return None
        if (
            self.filter.filter_units == FilterUnits.OBJECT_BOUNDING_BOX
        ):  # only percentages or fraction values
            bbox = self.bbox()
            north_west = bbox.north_west()
            if isinstance(self.filter.x, float):
                sx = self.filter.x
            else:
                sx = float(self.filter.x.rstrip("%")) / 100
            px = north_west.x + bbox.width * sx
            if isinstance(self.filter.y, float):
                sy = self.filter.y
            else:
                sy = float(self.filter.y.rstrip("%")) / 100
            py = north_west.y + bbox.height * sy
            if isinstance(self.filter.width, float):
                swidth = self.filter.width
            else:
                swidth = float(self.filter.width.rstrip("%")) / 100
            width = bbox.width * swidth
            if isinstance(self.filter.height, float):
                sheight = self.filter.height
            else:
                sheight = float(self.filter.height.rstrip("%")) / 100
            height = bbox.height * sheight
            filter_region = momapy.geometry.Bbox(
                momapy.geometry.Point(px + width / 2, py + height / 2),
                width,
                height,
            )
        else:  # only absolute values
            filter_region = momapy.geometry.Bbox(
                momapy.geometry.Point(
                    self.filter.x + self.filter.width / 2,
                    self.filter.y + self.filter.height / 2,
                ),
                self.filter.width,
                self.filter.height,
            )
        return filter_region


@dataclasses.dataclass(frozen=True, kw_only=True)
class Text(DrawingElement):
    """Class for text drawing elements"""

    text: str = dataclasses.field(
        metadata={"description": "The value of the text element"}
    )
    point: momapy.geometry.Point = dataclasses.field(
        metadata={"description": "The position of the text element"}
    )

    @property
    def x(self):
        return self.point.x

    @property
    def y(self):
        return self.point.y

    def transformed(
        self, transformation: momapy.geometry.Transformation
    ) -> typing_extensions.Self:
        """Return a copy of the text element"""
        return copy.deepcopy(self)

    def to_shapely(self, to_polygons=False) -> shapely.GeometryCollection:
        """Return a geometry collection built from the point of the text element"""
        return shapely.geometry.GeometryCollection([self.point.to_shapely()])


@dataclasses.dataclass(frozen=True, kw_only=True)
class Group(DrawingElement):
    """Class for group drawing elements"""

    elements: tuple[DrawingElement] = dataclasses.field(
        default_factory=tuple,
        metadata={"description": "The elements of the group element"},
    )

    def transformed(
        self, transformation: momapy.geometry.Transformation
    ) -> typing_extensions.Self:
        """Compute and return a copy of the group transformed with the given transformation"""
        elements = []
        for element in self.elements:
            elements.append(element.transformed(transformation))
        return dataclasses.replace(self, elements=tuple(elements))

    def to_shapely(self, to_polygons=False) -> shapely.GeometryCollection:
        """Return a geometry collection built from the elements of the group"""
        return drawing_elements_to_shapely(self.elements)


@dataclasses.dataclass(frozen=True)
class PathAction(abc.ABC):
    """Abstract class for path actions"""

    pass


@dataclasses.dataclass(frozen=True)
class MoveTo(PathAction):
    """Class for move-to path actions"""

    point: momapy.geometry.Point = dataclasses.field(
        metadata={"description": "The point of the move to action"}
    )

    @property
    def x(self):
        return self.point.x

    @property
    def y(self):
        return self.point.y

    def transformed(
        self,
        transformation: momapy.geometry.Transformation,
        current_point: momapy.geometry.Point,
    ) -> "MoveTo":
        """Compute and return a copy of the move-to path action transformed with the given transformation and current point"""
        return MoveTo(momapy.geometry.transform_point(self.point, transformation))


@dataclasses.dataclass(frozen=True)
class LineTo(PathAction):
    """Class for line-to path actions"""

    point: momapy.geometry.Point = dataclasses.field(
        metadata={"description": "The point of the line to action"}
    )

    @property
    def x(self):
        return self.point.x

    @property
    def y(self):
        return self.point.y

    def transformed(
        self,
        transformation: momapy.geometry.Transformation,
        current_point: momapy.geometry.Point,
    ):
        """Compute and return a copy of the line-to path action transformed with the given transformation and current point"""
        return LineTo(momapy.geometry.transform_point(self.point, transformation))

    def to_geometry(
        self, current_point: momapy.geometry.Point
    ) -> momapy.geometry.Segment:
        """Return a segment from the line-to path action and the given current point"""
        return momapy.geometry.Segment(current_point, self.point)

    def to_shapely(self, current_point: momapy.geometry.Point) -> shapely.LineString:
        """Return a line string from the line-to path action and the given current point"""
        return self.to_geometry(current_point).to_shapely()


@dataclasses.dataclass(frozen=True)
class EllipticalArc(PathAction):
    """Class for elliptical-arc path actions"""

    point: momapy.geometry.Point = dataclasses.field(
        metadata={"description": "The point of the elliptical arc action"}
    )
    rx: float = dataclasses.field(
        metadata={"description": "The x-radius of the elliptical arc action"}
    )
    ry: float = dataclasses.field(
        metadata={"description": "The y-radius of the elliptical arc action"}
    )
    x_axis_rotation: float = dataclasses.field(
        metadata={"description": "The x axis rotation of the elliptical arc action"}
    )
    arc_flag: int = dataclasses.field(
        metadata={"description": "The arc flag of the elliptical arc action"}
    )
    sweep_flag: int = dataclasses.field(
        metadata={"description": "The sweep flag of the elliptical arc action"}
    )

    @property
    def x(self):
        return self.point.x

    @property
    def y(self):
        return self.point.y

    def transformed(
        self,
        transformation: momapy.geometry.Transformation,
        current_point: momapy.geometry.Point,
    ) -> "EllipticalArc":
        """Compute and return a copy of the elliptical-arc path action transformed with the given transformation and current point"""
        east = momapy.geometry.Point(
            math.cos(self.x_axis_rotation) * self.rx,
            math.sin(self.x_axis_rotation) * self.rx,
        )
        north = momapy.geometry.Point(
            math.cos(self.x_axis_rotation) * self.ry,
            math.sin(self.x_axis_rotation) * self.ry,
        )
        new_center = momapy.geometry.transform_point(
            momapy.geometry.Point(0, 0), transformation
        )
        new_east = momapy.geometry.transform_point(east, transformation)
        new_north = momapy.geometry.transform_point(north, transformation)
        new_rx = momapy.geometry.Segment(new_center, new_east).length()
        new_ry = momapy.geometry.Segment(new_center, new_north).length()
        new_end_point = momapy.geometry.transform_point(self.point, transformation)
        new_x_axis_rotation = math.degrees(
            momapy.geometry.get_angle_to_horizontal_of_line(
                momapy.geometry.Line(new_center, new_east)
            )
        )
        return EllipticalArc(
            new_end_point,
            new_rx,
            new_ry,
            new_x_axis_rotation,
            self.arc_flag,
            self.sweep_flag,
        )

    def to_geometry(
        self, current_point: momapy.geometry.Point
    ) -> momapy.geometry.EllipticalArc:
        """Return an elliptical arc from the elliptical-arc path action and the given current point"""
        return momapy.geometry.EllipticalArc(
            current_point,
            self.point,
            self.rx,
            self.ry,
            self.x_axis_rotation,
            self.arc_flag,
            self.sweep_flag,
        )

    def to_shapely(
        self, current_point: momapy.geometry.Point
    ) -> shapely.GeometryCollection:
        """Return a line string reproducing the elliptical-arc path action from the given current point"""
        elliptical_arc = self.to_geometry(current_point)
        return elliptical_arc.to_shapely()


@dataclasses.dataclass(frozen=True)
class CurveTo(PathAction):
    """Class for curve-to path actions"""

    point: momapy.geometry.Point = dataclasses.field(
        metadata={"description": "The point of the curve to action"}
    )
    control_point1: momapy.geometry.Point = dataclasses.field(
        metadata={"description": "The first control point of the curve to action"}
    )
    control_point2: momapy.geometry.Point = dataclasses.field(
        metadata={"description": "The second control point of the curve to action"}
    )

    @property
    def x(self):
        return self.point.x

    @property
    def y(self):
        return self.point.y

    def transformed(
        self,
        transformation: momapy.geometry.Transformation,
        current_point: momapy.geometry.Point,
    ) -> "CurveTo":
        """Compute and return a copy of the curve-to path action transformed with the given transformation and current point"""
        return CurveTo(
            momapy.geometry.transform_point(self.point, transformation),
            momapy.geometry.transform_point(self.control_point1, transformation),
            momapy.geometry.transform_point(self.control_point2, transformation),
        )

    def to_geometry(
        self, current_point: momapy.geometry.Point
    ) -> momapy.geometry.BezierCurve:
        """Return a bezier curve from the curve-to path action and the given current point"""
        return momapy.geometry.BezierCurve(
            current_point,
            self.point,
            tuple([self.control_point1, self.control_point2]),
        )

    def to_shapely(
        self, current_point: momapy.geometry.Point, n_segs: int = 50
    ) -> shapely.GeometryCollection:
        """Return a line string reproducing the curve-to path action from the given current point"""
        bezier_curve = self.to_geometry(current_point)
        return bezier_curve.to_shapely(n_segs)


@dataclasses.dataclass(frozen=True)
class QuadraticCurveTo(PathAction):
    """Class for quadratic-curve-to paht actions"""

    point: momapy.geometry.Point = dataclasses.field(
        metadata={"description": "The point of the quadratic curve to action"}
    )
    control_point: momapy.geometry.Point = dataclasses.field(
        metadata={"description": "The control point of the quadratic curve to action"}
    )

    @property
    def x(self):
        return self.point.x

    @property
    def y(self):
        return self.point.y

    def transformed(
        self,
        transformation: momapy.geometry.Transformation,
        current_point: momapy.geometry.Point,
    ) -> "QuadraticCurveTo":
        """Compute and return a copy of the quadratic-curve-to transformed with the given transformation and current point"""
        return QuadraticCurveTo(
            momapy.geometry.transform_point(self.point, transformation),
            momapy.geometry.transform_point(self.control_point, transformation),
        )

    def to_curve_to(self, current_point: momapy.geometry.Point) -> CurveTo:
        """Return a curve-to path action repdroducing the quadratic-curve-to path action from the given current point"""
        p1 = current_point
        p2 = self.point
        control_point1 = p1 + (self.control_point - p1) * (2 / 3)
        control_point2 = p2 + (self.control_point - p2) * (2 / 3)
        return CurveTo(p2, control_point1, control_point2)

    def to_geometry(
        self, current_point: momapy.geometry.Point
    ) -> momapy.geometry.BezierCurve:
        """Return a bezier curve from the quadratic-curve-to path action and the given current point"""
        return momapy.geometry.BezierCurve(
            current_point,
            self.point,
            tuple([self.control_point]),
        )

    def to_shapely(self, current_point, n_segs=50):
        """Return a line string reproducing the quadratic-curve-to from the given current point"""
        bezier_curve = self.to_geometry(current_point)
        return bezier_curve.to_shapely()


@dataclasses.dataclass(frozen=True)
class ClosePath(PathAction):
    """Class for close-path path actions"""

    def transformed(
        self,
        transformation: momapy.geometry.Transformation,
        current_point: momapy.geometry.Point,
    ) -> "ClosePath":
        """Return a copy of the close-path path action"""
        return ClosePath()


@dataclasses.dataclass(frozen=True, kw_only=True)
class Path(DrawingElement):
    """Class for the path drawing elements"""

    actions: tuple[PathAction] = dataclasses.field(
        default_factory=tuple,
        metadata={"description": "The actions of the path"},
    )

    def transformed(
        self, transformation: momapy.geometry.Transformation
    ) -> typing_extensions.Self:
        """Compute and return a copy of the path element transformed with the given transformation"""
        actions = []
        current_point = None
        for action in self.actions:
            new_action = action.transformed(transformation, current_point)
            actions.append(new_action)
            if hasattr(action, "point"):
                current_point = action.point
            else:
                current_point = None
        return dataclasses.replace(self, actions=tuple(actions))

    def to_shapely(self, to_polygons=False) -> shapely.GeometryCollection:
        """Compute and return a geometry collection reproducing the path element"""
        current_point = momapy.geometry.Point(
            0, 0
        )  # in case the path does not start with a move_to command;
        # this is not possible in svg but not enforced here
        initial_point = current_point
        geom_collection = []
        line_strings = []
        previous_action = None
        for current_action in self.actions:
            if isinstance(current_action, MoveTo):
                current_point = current_action.point
                initial_point = current_point
                if (
                    not isinstance(previous_action, ClosePath)
                    and previous_action is not None
                ):
                    multi_linestring = shapely.geometry.MultiLineString(line_strings)
                    line_string = shapely.ops.linemerge(multi_linestring)
                    geom_collection.append(line_string)
                line_strings = []
            elif isinstance(current_action, ClosePath):
                if (
                    current_point.x != initial_point.x
                    or current_point.y != initial_point.y
                ):
                    line_string = shapely.geometry.LineString(
                        [current_point.to_tuple(), initial_point.to_tuple()]
                    )
                    line_strings.append(line_string)
                if not to_polygons:
                    multi_line = shapely.MultiLineString(line_strings)
                    line_string = shapely.ops.linemerge(multi_line)
                    geom_collection.append(line_string)
                else:
                    polygons = shapely.ops.polygonize(line_strings)
                    for polygon in polygons:
                        geom_collection.append(polygon)
                current_point = initial_point
            else:
                line_string = current_action.to_shapely(current_point)
                line_strings.append(line_string)
                current_point = current_action.point
            previous_action = current_action
        if not isinstance(current_action, (MoveTo, ClosePath)):
            multi_linestring = shapely.geometry.MultiLineString(line_strings)
            line_string = shapely.ops.linemerge(multi_linestring)
            geom_collection.append(line_string)
        return shapely.geometry.GeometryCollection(geom_collection)

    def to_path_with_bezier_curves(self) -> typing_extensions.Self:
        """Compute and return a path element reproducing the path element with curves and ellipses transformed to bezier curves"""
        new_actions = []
        current_point = momapy.geometry.Point(
            0, 0
        )  # in case the path does not start with a move_to command;
        # this is not possible in svg but not enforced here
        initial_point = current_point
        for action in self.actions:
            if isinstance(action, MoveTo):
                current_point = action.point
                initial_point = current_point
                new_actions.append(action)
            elif isinstance(action, ClosePath):
                current_point = initial_point
                new_actions.append(action)
            elif isinstance(
                action,
                (
                    LineTo,
                    CurveTo,
                    QuadraticCurveTo,
                ),
            ):
                current_point = action.point
                new_actions.append(action)
            elif isinstance(action, EllipticalArc):
                geom_object = action.to_geometry(current_point)
                bezier_curves = geom_object.to_bezier_curves()
                for bezier_curve in bezier_curves:
                    new_action = CurveTo(
                        point=bezier_curve.p2,
                        control_point1=bezier_curve.control_points[0],
                        control_point2=bezier_curve.control_points[1],
                    )
                    new_actions.append(new_action)
        new_path = dataclasses.replace(self, actions=new_actions)
        return new_path


@dataclasses.dataclass(frozen=True, kw_only=True)
class Ellipse(DrawingElement):
    """Class for ellipse drawing elements"""

    point: momapy.geometry.Point = dataclasses.field(
        metadata={"description": "The point of the ellipse"}
    )
    rx: float = dataclasses.field(
        metadata={"description": "The x-radius of the ellipse"}
    )
    ry: float = dataclasses.field(
        metadata={"description": "The y-radius of the ellipse"}
    )

    @property
    def x(self):
        return self.point.x

    @property
    def y(self):
        return self.point.y

    def to_path(self) -> Path:
        """Return a path element reproducing the ellipse element"""
        north = self.point + (0, -self.ry)
        east = self.point + (self.rx, 0)
        south = self.point + (0, self.ry)
        west = self.point - (self.rx, 0)
        actions = [
            MoveTo(north),
            EllipticalArc(east, self.rx, self.ry, 0, 0, 1),
            EllipticalArc(south, self.rx, self.ry, 0, 0, 1),
            EllipticalArc(west, self.rx, self.ry, 0, 0, 1),
            EllipticalArc(north, self.rx, self.ry, 0, 0, 1),
            ClosePath(),
        ]
        path = Path(
            stroke_width=self.stroke_width,
            stroke=self.stroke,
            fill=self.fill,
            transform=self.transform,
            filter=self.filter,
            actions=actions,
        )

        return path

    def transformed(self, transformation) -> Path:
        """Compute and return a copy of the ellipse element transformed with the given transformation"""
        path = self.to_path()
        return path.transformed(transformation)

    def to_shapely(self, to_polygons=False):
        """Return a geometry collection reproducing the ellipse element"""
        point = self.point.to_shapely()
        circle = point.buffer(1)
        ellipse = shapely.affinity.scale(circle, self.rx, self.ry)
        if not to_polygons:
            ellipse = ellipse.boundary
        return shapely.geometry.GeometryCollection([ellipse])


@dataclasses.dataclass(frozen=True, kw_only=True)
class Rectangle(DrawingElement):
    """Class for rectangle drawing elements"""

    point: momapy.geometry.Point = dataclasses.field(
        metadata={"description": "The point of the rectangle"}
    )
    width: float = dataclasses.field(
        metadata={"description": "The width of the rectangle"}
    )
    height: float = dataclasses.field(
        metadata={"description": "The height of the rectangle"}
    )
    rx: float = dataclasses.field(
        metadata={"description": "The x-radius of the rounded corners of the rectangle"}
    )
    ry: float = dataclasses.field(
        metadata={"description": "The y-radius of the rounded corners of the rectangle"}
    )

    @property
    def x(self):
        return self.point.x

    @property
    def y(self):
        return self.point.y

    def to_path(self) -> Path:
        """Return a path element reproducing the rectangle element"""
        x = self.point.x
        y = self.point.y
        rx = self.rx
        ry = self.ry
        width = self.width
        height = self.height
        actions = [
            MoveTo(momapy.geometry.Point(x + rx, y)),
            LineTo(momapy.geometry.Point(x + width - rx, y)),
        ]
        if rx > 0 and ry > 0:
            actions.append(
                EllipticalArc(momapy.geometry.Point(x + width, y + ry), rx, ry, 0, 0, 1)
            )
        actions.append(LineTo(momapy.geometry.Point(x + width, y + height - ry)))
        if rx > 0 and ry > 0:
            actions.append(
                EllipticalArc(
                    momapy.geometry.Point(x + width - rx, y + height),
                    rx,
                    ry,
                    0,
                    0,
                    1,
                )
            )
        actions.append(LineTo(momapy.geometry.Point(x + rx, y + height)))
        if rx > 0 and ry > 0:
            actions.append(
                EllipticalArc(
                    momapy.geometry.Point(x, y + height - ry), rx, ry, 0, 0, 1
                )
            )
        actions.append(LineTo(momapy.geometry.Point(x, y + ry)))
        if rx > 0 and ry > 0:
            actions.append(
                EllipticalArc(momapy.geometry.Point(x + rx, y), rx, ry, 0, 0, 1)
            )
        actions.append(ClosePath())
        path = Path(
            stroke_width=self.stroke_width,
            stroke=self.stroke,
            fill=self.fill,
            transform=self.transform,
            filter=self.filter,
            actions=actions,
        )

        return path

    def transformed(self, transformation: momapy.geometry.Transformation) -> Path:
        """Compute and return a copy of the rectangle element transformed with the given transformation"""
        path = self.to_path()
        return path.transformed(transformation)

    def to_shapely(self, to_polygons=False) -> shapely.GeometryCollection:
        """Return a geometry collection reproducing the rectangle element"""
        return self.to_path().to_shapely(to_polygons=to_polygons)


def drawing_elements_to_shapely(
    drawing_elements: collections.abc.Sequence[DrawingElement],
) -> shapely.GeometryCollection:
    """Return a geometry collection reproducing the given drawing elements"""
    geom_collection = []
    for drawing_element in drawing_elements:
        geom_collection += drawing_element.to_shapely(to_polygons=False).geoms
    return shapely.GeometryCollection(geom_collection)


def get_drawing_elements_border(
    drawing_elements: collections.abc.Sequence[DrawingElement],
    point: momapy.geometry.Point,
    center: momapy.geometry.Point | None = None,
) -> momapy.geometry.Point:
    """Compute and return the point at the intersection of the given drawing elements and the line passing through the given point and the center of the given drawing elements"""
    shapely_object = drawing_elements_to_shapely(drawing_elements)
    return momapy.geometry.get_shapely_object_border(
        shapely_object=shapely_object, point=point, center=center
    )


def get_drawing_elements_angle(
    drawing_elements: collections.abc.Sequence[DrawingElement],
    angle: float,
    unit="degrees",
    center: momapy.geometry.Point | None = None,
) -> momapy.geometry.Point:
    """Compute and return the point at the intersection of the given drawing elements and the line passing through the and the center of the given drawing elements and at the given angle from the horizontal"""
    shapely_object = drawing_elements_to_shapely(drawing_elements)
    return momapy.geometry.get_shapely_object_angle(
        shapely_object=shapely_object, angle=angle, unit=unit, center=center
    )


def get_drawing_elements_bbox(
    drawing_elements: collections.abc.Sequence[DrawingElement],
) -> momapy.geometry.Bbox:
    """Return the bounding box of the given drawing elements"""
    shapely_object = drawing_elements_to_shapely(drawing_elements)
    return momapy.geometry.Bbox.from_bounds(shapely_object.bounds)


def get_drawing_elements_anchor_point(
    drawing_elements,
    anchor_point: str,
    center: momapy.geometry.Point | None = None,
) -> momapy.geometry.Point:
    """Return the anchor point of the drawing elements"""
    shapely_object = drawing_elements_to_shapely(drawing_elements)
    return momapy.geometry.get_shapely_object_anchor_point(
        shapely_object=shapely_object, anchor_point=anchor_point, center=center
    )



from dataclasses import dataclass

import momapy.core
import momapy.meta.shapes
import momapy.drawing
import momapy.geometry


@dataclass(frozen=True, kw_only=True)
class PolyLine(momapy.core.SingleHeadedArc):
    """Class for single-headed arcs with no arrowhead"""

    def _arrowhead_border_drawing_elements(self):
        return []


@dataclass(frozen=True, kw_only=True)
class Triangle(momapy.core.SingleHeadedArc):
    """Class for single-headed arcs with a triangle arrowhead"""

    arrowhead_width: float = 10.0
    arrowhead_height: float = 10.0

    def _arrowhead_border_drawing_elements(self):
        shape = momapy.meta.shapes.Triangle(
            position=momapy.geometry.Point(self.arrowhead_width / 2, 0),
            width=self.arrowhead_width,
            height=self.arrowhead_height,
            direction=momapy.core.Direction.RIGHT,
        )
        return shape.drawing_elements()


@dataclass(frozen=True, kw_only=True)
class ReversedTriangle(momapy.core.SingleHeadedArc):
    """Class for single-headed arcs with a reversed triangle arrowhead"""

    arrowhead_width: float = 10.0
    arrowhead_height: float = 10.0

    def _arrowhead_border_drawing_elements(self):
        shape = momapy.meta.shapes.Triangle(
            position=momapy.geometry.Point(self.arrowhead_width / 2, 0),
            width=self.arrowhead_width,
            height=self.arrowhead_height,
            direction=momapy.core.Direction.LEFT,
        )
        return shape.drawing_elements()


@dataclass(frozen=True, kw_only=True)
class Rectangle(momapy.core.SingleHeadedArc):
    """Class for single-headed arcs with a rectangle arrowhead"""

    arrowhead_width: float = 10.0
    arrowhead_height: float = 10.0

    def _arrowhead_border_drawing_elements(self):
        shape = momapy.meta.shapes.Rectangle(
            position=momapy.geometry.Point(self.arrowhead_width / 2, 0),
            width=self.arrowhead_width,
            height=self.arrowhead_height,
        )
        return shape.drawing_elements()


@dataclass(frozen=True, kw_only=True)
class Ellipse(momapy.core.SingleHeadedArc):
    """Class for single-headed arcs with an ellipse arrowhead"""

    arrowhead_width: float = 10.0
    arrowhead_height: float = 5.0

    def _arrowhead_border_drawing_elements(self):
        shape = momapy.meta.shapes.Ellipse(
            position=momapy.geometry.Point(self.arrowhead_width / 2, 0),
            width=self.arrowhead_width,
            height=self.arrowhead_height,
        )
        return shape.drawing_elements()


@dataclass(frozen=True, kw_only=True)
class Diamond(momapy.core.SingleHeadedArc):
    """Class for single-headed arcs with a diamond arrowhead"""

    arrowhead_width: float = 10.0
    arrowhead_height: float = 5.0

    def _arrowhead_border_drawing_elements(self):
        shape = momapy.meta.shapes.Diamond(
            position=momapy.geometry.Point(self.arrowhead_width / 2, 0),
            width=self.arrowhead_width,
            height=self.arrowhead_height,
        )
        return shape.drawing_elements()


@dataclass(frozen=True, kw_only=True)
class Bar(momapy.core.SingleHeadedArc):
    """Class for single-headed arcs with a bar arrowhead"""

    arrowhead_height: float = 10.0

    def _arrowhead_border_drawing_elements(self):
        shape = momapy.meta.shapes.Bar(
            position=momapy.geometry.Point(0, 0),
            height=self.arrowhead_height,
        )
        return shape.drawing_elements()


@dataclass(frozen=True, kw_only=True)
class ArcBarb(momapy.core.SingleHeadedArc):
    """Class for single-headed arcs with an arc barb arrowhead"""

    arrowhead_width: float = 5.0
    arrowhead_height: float = 10.0
    arrowhead_fill: (
        momapy.drawing.NoneValueType | momapy.coloring.Color | None
    ) = momapy.drawing.NoneValue

    def _arrowhead_border_drawing_elements(self):
        shape = momapy.meta.shapes.ArcBarb(
            position=momapy.geometry.Point(-self.arrowhead_width / 2, 0),
            width=self.arrowhead_width,
            height=self.arrowhead_height,
            direction=momapy.core.Direction.RIGHT,
        )
        return shape.drawing_elements()


@dataclass(frozen=True, kw_only=True)
class StraightBarb(momapy.core.SingleHeadedArc):
    """Class for single-headed arcs with a straight barb arrowhead"""

    arrowhead_width: float = 10.0
    arrowhead_height: float = 10.0
    arrowhead_fill: (
        momapy.drawing.NoneValueType | momapy.coloring.Color | None
    ) = momapy.drawing.NoneValue

    def _arrowhead_border_drawing_elements(self):
        shape = momapy.meta.shapes.StraightBarb(
            position=momapy.geometry.Point(-self.arrowhead_width / 2, 0),
            width=self.arrowhead_width,
            height=self.arrowhead_height,
            direction=momapy.core.Direction.RIGHT,
        )
        return shape.drawing_elements()


@dataclass(frozen=True, kw_only=True)
class To(momapy.core.SingleHeadedArc):
    """Class for single-headed arcs with a to arrowhead"""

    arrowhead_width: float = 5.0
    arrowhead_height: float = 10.0
    arrowhead_fill: (
        momapy.drawing.NoneValueType | momapy.coloring.Color | None
    ) = momapy.drawing.NoneValue

    def _arrowhead_border_drawing_elements(self):
        shape = momapy.meta.shapes.To(
            position=momapy.geometry.Point(-self.arrowhead_width / 2, 0),
            width=self.arrowhead_width,
            height=self.arrowhead_height,
            direction=momapy.core.Direction.RIGHT,
        )
        return shape.drawing_elements()


@dataclass(frozen=True, kw_only=True)
class DoubleTriangle(momapy.core.DoubleHeadedArc):
    """Class for double-headed arcs with triangle arrowheads"""

    start_arrowhead_width: float = 10.0
    start_arrowhead_height: float = 10.0
    end_arrowhead_width: float = 10.0
    end_arrowhead_height: float = 10.0

    def _start_arrowhead_border_drawing_elements(self):
        shape = momapy.meta.shapes.Triangle(
            position=momapy.geometry.Point(-self.start_arrowhead_width / 2, 0),
            width=self.start_arrowhead_width,
            height=self.start_arrowhead_height,
            direction=momapy.core.Direction.LEFT,
        )
        return shape.drawing_elements()

    def _end_arrowhead_border_drawing_elements(self):
        shape = momapy.meta.shapes.Triangle(
            position=momapy.geometry.Point(self.end_arrowhead_width / 2, 0),
            width=self.end_arrowhead_width,
            height=self.end_arrowhead_height,
            direction=momapy.core.Direction.RIGHT,
        )
        return shape.drawing_elements()



import momapy.io.core

_registered = False


def _ensure_registered():
    global _registered
    if not _registered:
        import momapy.sbgn.io.sbgnml
        import momapy.sbgn.io.pickle
        import momapy.celldesigner.io.celldesigner
        import momapy.celldesigner.io.pickle
        import momapy.sbml.io.sbml

        momapy.io.core.register_reader(
            "sbgnml-0.2", momapy.sbgn.io.sbgnml.SBGNML0_2Reader
        )
        momapy.io.core.register_reader(
            "sbgnml-0.3", momapy.sbgn.io.sbgnml.SBGNML0_3Reader
        )
        momapy.io.core.register_reader("sbgnml", momapy.sbgn.io.sbgnml.SBGNML0_3Reader)
        momapy.io.core.register_writer(
            "sbgnml-0.3", momapy.sbgn.io.sbgnml.SBGNML0_3Writer
        )
        momapy.io.core.register_writer("sbgnml", momapy.sbgn.io.sbgnml.SBGNML0_3Writer)

        momapy.io.core.register_reader(
            "sbgn-pickle", momapy.sbgn.io.pickle.SBGNPickleReader
        )
        momapy.io.core.register_writer(
            "sbgn-pickle", momapy.sbgn.io.pickle.SBGNPickleWriter
        )

        momapy.io.core.register_reader(
            "celldesigner", momapy.celldesigner.io.celldesigner.CellDesignerReader
        )
        momapy.io.core.register_reader(
            "celldesigner-pickle",
            momapy.celldesigner.io.pickle.CellDesignerPickleReader,
        )
        momapy.io.core.register_writer(
            "celldesigner-pickle",
            momapy.celldesigner.io.pickle.CellDesignerPickleWriter,
        )

        momapy.io.core.register_reader("sbml", momapy.sbml.io.sbml.SBMLReader)
        _registered = True



"""Classes for common node types"""

import dataclasses

import momapy.core
import momapy.meta.shapes


@dataclasses.dataclass(frozen=True, kw_only=True)
class Rectangle(momapy.core.Node):
    """Class for rectangle nodes"""

    top_left_rx: float = 0.0
    top_left_ry: float = 0.0
    top_left_rounded_or_cut: str = "rounded"
    top_right_rx: float = 0.0
    top_right_ry: float = 0.0
    top_right_rounded_or_cut: str = "rounded"
    bottom_right_rx: float = 0.0
    bottom_right_ry: float = 0.0
    bottom_right_rounded_or_cut: str = "rounded"
    bottom_left_rx: float = 0.0
    bottom_left_ry: float = 0.0
    bottom_left_rounded_or_cut: str = "rounded"

    def _border_drawing_elements(self):
        shape = momapy.meta.shapes.Rectangle(
            position=self.position,
            width=self.width,
            height=self.height,
            top_left_rx=self.top_left_rx,
            top_left_ry=self.top_left_ry,
            top_left_rounded_or_cut=self.top_left_rounded_or_cut,
            top_right_rx=self.top_right_rx,
            top_right_ry=self.top_right_ry,
            top_right_rounded_or_cut=self.top_right_rounded_or_cut,
            bottom_left_rx=self.bottom_left_rx,
            bottom_left_ry=self.bottom_left_ry,
            bottom_left_rounded_or_cut=self.bottom_left_rounded_or_cut,
            bottom_right_rx=self.bottom_right_rx,
            bottom_right_ry=self.bottom_right_ry,
            bottom_right_rounded_or_cut=self.bottom_right_rounded_or_cut,
        )
        return shape.drawing_elements()


@dataclasses.dataclass(frozen=True, kw_only=True)
class Ellipse(momapy.core.Node):
    """Class for ellipse nodes"""

    def _border_drawing_elements(self):
        shape = momapy.meta.shapes.Ellipse(
            position=self.position,
            width=self.width,
            height=self.height,
        )
        return shape.drawing_elements()


@dataclasses.dataclass(frozen=True, kw_only=True)
class Stadium(momapy.core.Node):
    """Class for stadium nodes"""

    def _border_drawing_elements(self):
        shape = momapy.meta.shapes.Stadium(
            position=self.position,
            width=self.width,
            height=self.height,
        )
        return shape.drawing_elements()


@dataclasses.dataclass(frozen=True, kw_only=True)
class Hexagon(momapy.core.Node):
    """Class for hexagon nodes"""

    left_angle: float = 60.0
    right_angle: float = 60.0

    def _border_drawing_elements(self):
        shape = momapy.meta.shapes.Hexagon(
            position=self.position,
            width=self.width,
            height=self.height,
            left_angle=self.left_angle,
            right_angle=self.right_angle,
        )
        return shape.drawing_elements()


@dataclasses.dataclass(frozen=True, kw_only=True)
class TurnedHexagon(momapy.core.Node):
    """Class for hexagon turned by 90 degrees nodes"""

    top_angle: float = 80.0
    bottom_angle: float = 80.0

    def _border_drawing_elements(self):
        shape = momapy.meta.shapes.TurnedHexagon(
            position=self.position,
            width=self.width,
            height=self.height,
            top_angle=self.top_angle,
            bottom_angle=self.bottom_angle,
        )
        return shape.drawing_elements()


@dataclasses.dataclass(frozen=True, kw_only=True)
class Parallelogram(momapy.core.Node):
    """Class for parallelogram nodes"""

    angle: float = 60.0

    def _border_drawing_elements(self):
        shape = momapy.meta.shapes.Parallelogram(
            position=self.position,
            width=self.width,
            height=self.height,
            angle=self.angle,
        )
        return shape.drawing_elements()


@dataclasses.dataclass(frozen=True, kw_only=True)
class CrossPoint(momapy.core.Node):
    """Class for cross point nodes"""

    def _border_drawing_elements(self):
        shape = momapy.meta.shapes.CrossPoint(
            position=self.position,
            width=self.width,
            height=self.height,
        )
        return shape.drawing_elements()


@dataclasses.dataclass(frozen=True, kw_only=True)
class Triangle(momapy.core.Node):
    """Class for triangle nodes"""

    direction: momapy.core.Direction = momapy.core.Direction.RIGHT

    def _border_drawing_elements(self):
        shape = momapy.meta.shapes.Triangle(
            position=self.position,
            width=self.width,
            height=self.height,
            direction=self.direction,
        )
        return shape.drawing_elements()


@dataclasses.dataclass(frozen=True, kw_only=True)
class Diamond(momapy.core.Node):
    """Class for diamond nodes"""

    def _border_drawing_elements(self):
        shape = momapy.meta.shapes.Diamond(
            position=self.position,
            width=self.width,
            height=self.height,
        )
        return shape.drawing_elements()


@dataclasses.dataclass(frozen=True, kw_only=True)
class Bar(momapy.core.Node):
    """Class for bar nodes"""

    def _border_drawing_elements(self):
        shape = momapy.meta.shapes.Bar(
            position=self.position,
            height=self.height,
        )
        return shape.drawing_elements()


@dataclasses.dataclass(frozen=True, kw_only=True)
class ArcBarb(momapy.core.Node):
    """Class for arc barb nodes"""

    direction: momapy.core.Direction = momapy.core.Direction.RIGHT

    def _border_drawing_elements(self):
        shape = momapy.meta.shapes.ArcBarb(
            position=self.position,
            width=self.width,
            height=self.height,
            direction=self.direction,
        )
        return shape.drawing_elements()


@dataclasses.dataclass(frozen=True, kw_only=True)
class StraightBarb(momapy.core.Node):
    """Class for straight barb nodes"""

    direction: momapy.core.Direction = momapy.core.Direction.RIGHT

    def _border_drawing_elements(self):
        shape = momapy.meta.shapes.StraightBarb(
            position=self.position,
            width=self.width,
            height=self.height,
            direction=self.direction,
        )
        return shape.drawing_elements()


@dataclasses.dataclass(frozen=True, kw_only=True)
class To(momapy.core.Node):
    """Class for to nodes"""

    direction: momapy.core.Direction = momapy.core.Direction.RIGHT

    def _border_drawing_elements(self):
        shape = momapy.meta.shapes.To(
            position=self.position,
            width=self.width,
            height=self.height,
            direction=self.direction,
        )
        return shape.drawing_elements()



import math
import dataclasses

import momapy.core
import momapy.geometry
import momapy.drawing


@dataclasses.dataclass(frozen=True, kw_only=True)
class Rectangle(momapy.core.Shape):
    """Class for rectangle shapes"""

    position: momapy.geometry.Point
    width: float
    height: float
    top_left_rx: float = 0.0
    top_left_ry: float = 0.0
    top_left_rounded_or_cut: str = "rounded"
    top_right_rx: float = 0.0
    top_right_ry: float = 0.0
    top_right_rounded_or_cut: str = "rounded"
    bottom_right_rx: float = 0.0
    bottom_right_ry: float = 0.0
    bottom_right_rounded_or_cut: str = "rounded"
    bottom_left_rx: float = 0.0
    bottom_left_ry: float = 0.0
    bottom_left_rounded_or_cut: str = "rounded"

    def joint1(self):
        return self.position + (
            -self.width / 2 + self.top_left_rx,
            -self.height / 2,
        )

    def joint2(self):
        return self.position + (
            self.width / 2 - self.top_right_rx,
            -self.height / 2,
        )

    def joint3(self):
        return self.position + (
            self.width / 2,
            -self.height / 2 + self.top_right_ry,
        )

    def joint4(self):
        return self.position + (
            self.width / 2,
            self.height / 2 - self.bottom_right_ry,
        )

    def joint5(self):
        return self.position + (
            self.width / 2 - self.bottom_right_rx,
            self.height / 2,
        )

    def joint6(self):
        return self.position + (
            -self.width / 2 + self.bottom_left_rx,
            self.height / 2,
        )

    def joint7(self):
        return self.position + (
            -self.width / 2,
            self.height / 2 - self.bottom_left_ry,
        )

    def joint8(self):
        return self.position + (
            -self.width / 2,
            -self.height / 2 + self.top_left_ry,
        )

    def drawing_elements(self):
        actions = [
            momapy.drawing.MoveTo(self.joint1()),
            momapy.drawing.LineTo(self.joint2()),
        ]
        if self.top_right_rx != 0 and self.top_right_ry != 0:
            if self.top_right_rounded_or_cut == "cut":
                actions.append(momapy.drawing.LineTo(self.joint3()))
            elif self.top_right_rounded_or_cut == "rounded":
                actions.append(
                    momapy.drawing.EllipticalArc(
                        self.joint3(),
                        self.top_right_rx,
                        self.top_right_ry,
                        0,
                        0,
                        1,
                    )
                ),
        actions.append(momapy.drawing.LineTo(self.joint4()))
        if self.bottom_right_rx != 0 and self.bottom_right_ry != 0:
            if self.bottom_right_rounded_or_cut == "cut":
                actions.append(momapy.drawing.LineTo(self.joint5()))
            elif self.bottom_right_rounded_or_cut == "rounded":
                actions.append(
                    momapy.drawing.EllipticalArc(
                        self.joint5(),
                        self.bottom_right_rx,
                        self.bottom_right_ry,
                        0,
                        0,
                        1,
                    )
                ),
        actions.append(momapy.drawing.LineTo(self.joint6()))
        if self.bottom_left_rx != 0 and self.bottom_left_ry != 0:
            if self.bottom_left_rounded_or_cut == "cut":
                actions.append(momapy.drawing.LineTo(self.joint7()))
            elif self.bottom_left_rounded_or_cut == "rounded":
                actions.append(
                    momapy.drawing.EllipticalArc(
                        self.joint7(),
                        self.bottom_left_rx,
                        self.bottom_left_ry,
                        0,
                        0,
                        1,
                    )
                ),
        actions.append(momapy.drawing.LineTo(self.joint8()))
        if self.top_left_rx != 0 and self.top_left_ry != 0:
            if self.top_left_rounded_or_cut == "rounded":
                actions.append(
                    momapy.drawing.EllipticalArc(
                        self.joint1(),
                        self.top_left_rx,
                        self.top_left_ry,
                        0,
                        0,
                        1,
                    )
                ),
        actions.append(momapy.drawing.ClosePath())
        drawing_element = momapy.drawing.Path(actions=actions)
        return [drawing_element]


@dataclasses.dataclass(frozen=True, kw_only=True)
class Ellipse(momapy.core.Shape):
    """Class for ellipse shapes"""

    position: momapy.geometry.Point
    width: float
    height: float

    def drawing_elements(self):
        drawing_element = momapy.drawing.Ellipse(
            point=self.position, rx=self.width / 2, ry=self.height / 2
        )
        return [drawing_element]


@dataclasses.dataclass(frozen=True, kw_only=True)
class Stadium(momapy.core.Shape):
    """Class for stadium shapes"""

    position: momapy.geometry.Point
    width: float
    height: float

    def __post_init__(self):
        if self.width < self.height:
            object.__setattr__(self, "width", self.height)

    def joint1(self):
        return self.position + (
            self.height / 2 - self.width / 2,
            -self.height / 2,
        )

    def joint2(self):
        return self.position + (
            self.width / 2 - self.height / 2,
            -self.height / 2,
        )

    def joint3(self):
        return self.position + (
            self.width / 2 - self.height / 2,
            self.height / 2,
        )

    def joint4(self):
        return self.position + (
            self.height / 2 - self.width / 2,
            self.height / 2,
        )

    def drawing_elements(self):
        drawing_element = momapy.drawing.Rectangle(
            point=self.position - (self.width / 2, self.height / 2),
            height=self.height,
            width=self.width,
            rx=self.height / 2,
            ry=self.height / 2,
        )
        return [drawing_element]


@dataclasses.dataclass(frozen=True, kw_only=True)
class Hexagon(momapy.core.Shape):
    """Class for hexagon shapes"""

    position: momapy.geometry.Point
    width: float
    height: float
    left_angle: float
    right_angle: float

    def joint1(self):
        if self.left_angle > 90:
            return self.position + (-self.width / 2, -self.height / 2)
        angle = math.radians(self.left_angle)
        side_length = abs(self.height / (math.tan(angle)))
        return self.position + (
            -self.width / 2 + side_length,
            -self.height / 2,
        )

    def joint2(self):
        if self.right_angle > 90:
            return self.position + (self.width / 2, -self.height / 2)
        angle = math.radians(self.right_angle)
        side_length = abs(self.height / (math.tan(angle)))
        return self.position + (self.width / 2 - side_length, -self.height / 2)

    def joint3(self):
        if self.right_angle <= 90:
            return self.position + (self.width / 2, 0)
        angle = math.radians(180 - self.right_angle)
        side_length = abs(self.height / (math.tan(angle)))
        return self.position + (self.width / 2 - side_length, 0)

    def joint4(self):
        if self.right_angle > 90:
            return self.position + (self.width / 2, self.height / 2)
        angle = math.radians(self.right_angle)
        side_length = abs(self.height / (math.tan(angle)))
        return self.position + (self.width / 2 - side_length, self.height / 2)

    def joint5(self):
        if self.left_angle > 90:
            return self.position + (-self.width / 2, self.height / 2)
        angle = math.radians(self.left_angle)
        side_length = abs(self.height / (math.tan(angle)))
        return self.position + (-self.width / 2 + side_length, self.height / 2)

    def joint6(self):
        if self.left_angle <= 90:
            return self.position + (-self.width / 2, 0)
        angle = math.radians(180 - self.left_angle)
        side_length = abs(self.height / (math.tan(angle)))
        return self.position + (-self.width / 2 + side_length, 0)

    def drawing_elements(self):
        actions = [
            momapy.drawing.MoveTo(self.joint1()),
            momapy.drawing.LineTo(self.joint2()),
            momapy.drawing.LineTo(self.joint3()),
            momapy.drawing.LineTo(self.joint4()),
            momapy.drawing.LineTo(self.joint5()),
            momapy.drawing.LineTo(self.joint6()),
            momapy.drawing.ClosePath(),
        ]
        drawing_element = momapy.drawing.Path(actions=actions)
        return [drawing_element]


@dataclasses.dataclass(frozen=True, kw_only=True)
class TurnedHexagon(momapy.core.Shape):
    """Class for hexagon turned by 90 degrees shapes"""

    position: momapy.geometry.Point
    width: float
    height: float
    top_angle: float
    bottom_angle: float

    def joint1(self):
        if self.top_angle >= 90:
            p = self.position + (-self.width / 2, -self.height / 2)
        else:
            p = self.position + (0, -self.height / 2)
        return p

    def joint2(self):
        if self.top_angle >= 90:
            angle = math.radians(180 - self.top_angle)
            side_length = abs(self.width / (math.tan(angle)))
            p = self.position + (0, -self.height / 2 + side_length)
        else:
            angle = math.radians(self.top_angle)
            side_length = abs(self.width / (math.tan(angle)))
            p = self.position + (
                self.width / 2,
                -self.height / 2 + side_length,
            )
        return p

    def joint3(self):
        if self.top_angle >= 90:
            p = self.position + (self.width / 2, -self.height / 2)
        else:
            if self.bottom_angle >= 90:
                p = self.position + (self.width / 2, self.height / 2)
            else:
                angle = math.radians(self.top_angle)
                side_length = abs(self.width / (math.tan(angle)))
                p = self.position + (
                    self.width / 2,
                    self.height / 2 - side_length,
                )
        return p

    def joint4(self):
        if self.top_angle >= 90:
            if self.bottom_angle >= 90:
                p = self.position + (self.width / 2, self.height / 2)
            else:
                angle = math.radians(self.bottom_angle)
                side_length = abs(self.width / (math.tan(angle)))
                p = self.position + (
                    self.width / 2,
                    self.height / 2 - side_length,
                )
        else:
            if self.bottom_angle >= 90:
                angle = math.radians(180 - self.bottom_angle)
                side_length = abs(self.width / (math.tan(angle)))
                p = self.position + (
                    0,
                    self.height / 2 - side_length,
                )
            else:
                p = self.position + (0, self.height / 2)

        return p

    def joint5(self):
        if self.top_angle >= 90:
            if self.bottom_angle >= 90:
                angle = math.radians(180 - self.top_angle)
                side_length = abs(self.width / (math.tan(angle)))
                p = self.position + (0, self.height / 2 - side_length)
            else:
                p = self.position + (0, self.height / 2)
        else:
            if self.bottom_angle >= 90:
                p = self.position + (-self.width / 2, self.height / 2)
            else:
                angle = math.radians(self.bottom_angle)
                side_length = abs(self.width / (math.tan(angle)))
                p = self.position + (
                    -self.width / 2,
                    self.height / 2 - side_length,
                )
        return p

    def joint6(self):
        if self.top_angle >= 90:
            if self.bottom_angle >= 90:
                return self.position + (-self.width / 2, self.height / 2)
            else:
                angle = math.radians(180 - self.bottom_angle)
                side_length = abs(self.width / (math.tan(angle)))
                p = self.position + (
                    -self.width / 2,
                    self.height / 2 - side_length,
                )
        else:
            angle = math.radians(self.top_angle)
            side_length = abs(self.width / (math.tan(angle)))
            p = self.position + (
                -self.width / 2,
                -self.height / 2 + side_length,
            )
        return p

    def drawing_elements(self):
        actions = [
            momapy.drawing.MoveTo(self.joint1()),
            momapy.drawing.LineTo(self.joint2()),
            momapy.drawing.LineTo(self.joint3()),
            momapy.drawing.LineTo(self.joint4()),
            momapy.drawing.LineTo(self.joint5()),
            momapy.drawing.LineTo(self.joint6()),
            momapy.drawing.ClosePath(),
        ]
        drawing_element = momapy.drawing.Path(actions=actions)
        return [drawing_element]


@dataclasses.dataclass(frozen=True, kw_only=True)
class Parallelogram(momapy.core.Shape):
    """Class for parallelogram shapes"""

    position: momapy.geometry.Point
    width: float
    height: float
    angle: float

    def joint1(self):
        if self.angle >= 90:
            return self.position + (-self.width / 2, -self.height / 2)
        angle = math.radians(self.angle)
        side_length = abs(self.height / math.tan(angle))
        return self.position + (
            -self.width / 2 + side_length,
            -self.height / 2,
        )

    def joint2(self):
        if self.angle <= 90:
            return self.position + (self.width / 2, -self.height / 2)
        angle = math.radians(180 - self.angle)
        side_length = abs(self.height / math.tan(angle))
        return self.position + (self.width / 2 - side_length, -self.height / 2)

    def joint3(self):
        if self.angle > 90:
            return self.position + (self.width / 2, self.height / 2)
        angle = math.radians(self.angle)
        side_length = abs(self.height / math.tan(angle))
        return self.position + (self.width / 2 - side_length, self.height / 2)

    def joint4(self):
        if self.angle <= 90:
            return self.position + (-self.width / 2, self.height / 2)
        angle = math.radians(180 - self.angle)
        side_length = abs(self.height / math.tan(angle))
        return self.position + (-self.width / 2 + side_length, self.height / 2)

    def drawing_elements(self):
        actions = [
            momapy.drawing.MoveTo(self.joint1()),
            momapy.drawing.LineTo(self.joint2()),
            momapy.drawing.LineTo(self.joint3()),
            momapy.drawing.LineTo(self.joint4()),
            momapy.drawing.ClosePath(),
        ]
        drawing_element = momapy.drawing.Path(actions=actions)
        return [drawing_element]


@dataclasses.dataclass(frozen=True, kw_only=True)
class CrossPoint(momapy.core.Shape):
    """Class for cross point shapes"""

    position: momapy.geometry.Point
    width: float
    height: float

    def drawing_elements(self):
        actions = [
            momapy.drawing.MoveTo(self.position - (self.width / 2, 0)),
            momapy.drawing.LineTo(self.position + (self.width / 2, 0)),
        ]
        horizontal_path = momapy.drawing.Path(actions=actions)
        actions = [
            momapy.drawing.MoveTo(self.position - (0, self.height / 2, 0)),
            momapy.drawing.LineTo(self.position + (0, self.height / 2)),
        ]
        vertical_path = momapy.drawing.Path(actions=actions)
        return [horizontal_path, vertical_path]


@dataclasses.dataclass(frozen=True, kw_only=True)
class Triangle(momapy.core.Shape):
    """Class for triangle shapes"""

    position: momapy.geometry.Point
    width: float
    height: float
    direction: momapy.core.Direction

    def joint1(self):
        if (
            self.direction == momapy.core.Direction.RIGHT
            or self.direction == momapy.core.Direction.DOWN
        ):
            p = self.position + (-self.width / 2, -self.height / 2)
        elif self.direction == momapy.core.Direction.UP:
            p = self.position + (0, -self.height / 2)
        elif self.direction == momapy.core.Direction.LEFT:
            p = self.position + (self.width / 2, -self.height / 2)
        return p

    def joint2(self):
        if self.direction == momapy.core.Direction.RIGHT:
            p = self.position + (self.width / 2, 0)
        elif (
            self.direction == momapy.core.Direction.UP
            or self.direction == momapy.core.Direction.LEFT
        ):
            p = self.position + (self.width / 2, self.height / 2)
        elif self.direction == momapy.core.Direction.DOWN:
            p = self.position + (self.width / 2, -self.height / 2)
        return p

    def joint3(self):
        if (
            self.direction == momapy.core.Direction.RIGHT
            or self.direction == momapy.core.Direction.UP
        ):
            p = self.position + (-self.width / 2, self.height / 2)
        elif self.direction == momapy.core.Direction.LEFT:
            p = self.position + (-self.width / 2, 0)
        elif self.direction == momapy.core.Direction.DOWN:
            p = self.position + (0, self.height / 2)
        return p

    def drawing_elements(self):
        actions = [
            momapy.drawing.MoveTo(self.joint1()),
            momapy.drawing.LineTo(self.joint2()),
            momapy.drawing.LineTo(self.joint3()),
            momapy.drawing.ClosePath(),
        ]
        drawing_element = momapy.drawing.Path(actions=actions)
        return [drawing_element]


@dataclasses.dataclass(frozen=True, kw_only=True)
class Diamond(momapy.core.Shape):
    """Class for diamond shapes"""

    position: momapy.geometry.Point
    width: float
    height: float

    def joint1(self):
        return self.position + (0, -self.height / 2)

    def joint2(self):
        return self.position + (self.width / 2, 0)

    def joint3(self):
        return self.position + (0, self.height / 2)

    def joint4(self):
        return self.position + (-self.width / 2, 0)

    def drawing_elements(self):
        actions = [
            momapy.drawing.MoveTo(self.joint1()),
            momapy.drawing.LineTo(self.joint2()),
            momapy.drawing.LineTo(self.joint3()),
            momapy.drawing.LineTo(self.joint4()),
            momapy.drawing.ClosePath(),
        ]
        drawing_element = momapy.drawing.Path(actions=actions)
        return [drawing_element]


@dataclasses.dataclass(frozen=True, kw_only=True)
class Bar(momapy.core.Shape):
    """Class for bar shapes"""

    position: momapy.geometry.Point
    height: float

    def joint1(self):
        return momapy.geometry.Point(0, -self.height / 2)

    def joint2(self):
        return momapy.geometry.Point(0, self.height / 2)

    def drawing_elements(self):
        actions = [
            momapy.drawing.MoveTo(self.joint1()),
            momapy.drawing.LineTo(self.joint2()),
        ]
        drawing_element = momapy.drawing.Path(actions=actions)
        return [drawing_element]


@dataclasses.dataclass(frozen=True, kw_only=True)
class ArcBarb(momapy.core.Shape):
    """Class for arc barb shapes"""

    position: momapy.geometry.Point
    width: float
    height: float
    direction: momapy.core.Direction

    def joint1(self):
        if (
            self.direction == momapy.core.Direction.RIGHT
            or self.direction == momapy.core.Direction.DOWN
        ):
            p = self.position + (-self.width / 2, -self.height / 2)
        elif self.direction == momapy.core.Direction.UP:
            p = self.position + (self.width / 2, self.height / 2)
        elif self.direction == momapy.core.Direction.LEFT:
            p = self.position + (self.width / 2, -self.height / 2)
        return p

    def joint2(self):
        if (
            self.direction == momapy.core.Direction.RIGHT
            or self.direction == momapy.core.Direction.UP
        ):
            p = self.position + (-self.width / 2, self.height / 2)
        elif self.direction == momapy.core.Direction.LEFT:
            p = self.position + (self.width / 2, self.height / 2)
        elif self.direction == momapy.core.Direction.DOWN:
            p = self.position + (self.width / 2, -self.height / 2)
        return p

    def drawing_elements(self):
        if self.direction == momapy.core.Direction.RIGHT:
            actions = [
                momapy.drawing.MoveTo(self.joint1()),
                momapy.drawing.EllipticalArc(
                    self.joint2(),
                    self.width,
                    self.height / 2,
                    0,
                    0,
                    1,
                ),
            ]
        elif self.direction == momapy.core.Direction.UP:
            actions = [
                momapy.drawing.MoveTo(self.joint1()),
                momapy.drawing.EllipticalArc(
                    self.joint2(),
                    self.width / 2,
                    self.height,
                    0,
                    0,
                    0,
                ),
            ]
        elif self.direction == momapy.core.Direction.LEFT:
            actions = [
                momapy.drawing.MoveTo(self.joint1()),
                momapy.drawing.EllipticalArc(
                    self.joint2(),
                    self.width,
                    self.height / 2,
                    0,
                    0,
                    0,
                ),
            ]
        elif self.direction == momapy.core.Direction.DOWN:
            actions = [
                momapy.drawing.MoveTo(self.joint1()),
                momapy.drawing.EllipticalArc(
                    self.joint2(),
                    self.width / 2,
                    self.height,
                    0,
                    0,
                    0,
                ),
            ]
        drawing_element = momapy.drawing.Path(actions=actions)
        return [drawing_element]


@dataclasses.dataclass(frozen=True, kw_only=True)
class StraightBarb(momapy.core.Shape):
    """Class for straight barb shapes"""

    position: momapy.geometry.Point
    width: float
    height: float
    direction: momapy.core.Direction

    def joint1(self):
        if (
            self.direction == momapy.core.Direction.RIGHT
            or self.direction == momapy.core.Direction.DOWN
        ):
            p = self.position + (-self.width / 2, -self.height / 2)
        elif self.direction == momapy.core.Direction.UP:
            p = self.position + (0, -self.height / 2)
        elif self.direction == momapy.core.Direction.LEFT:
            p = self.position + (self.width / 2, -self.height / 2)
        return p

    def joint2(self):
        if self.direction == momapy.core.Direction.RIGHT:
            p = self.position + (self.width / 2, 0)
        elif (
            self.direction == momapy.core.Direction.UP
            or self.direction == momapy.core.Direction.LEFT
        ):
            p = self.position + (self.width / 2, self.height / 2)
        elif self.direction == momapy.core.Direction.DOWN:
            p = self.position + (self.width / 2, -self.height / 2)
        return p

    def joint3(self):
        if (
            self.direction == momapy.core.Direction.RIGHT
            or self.direction == momapy.core.Direction.UP
        ):
            p = self.position + (-self.width / 2, self.height / 2)
        elif self.direction == momapy.core.Direction.LEFT:
            p = self.position + (-self.width / 2, 0)
        elif self.direction == momapy.core.Direction.DOWN:
            p = self.position + (0, self.height / 2)
        return p

    def drawing_elements(self):
        if self.direction == momapy.core.Direction.RIGHT:
            actions = [
                momapy.drawing.MoveTo(self.joint1()),
                momapy.drawing.LineTo(self.joint2()),
                momapy.drawing.LineTo(self.joint3()),
            ]
        elif self.direction == momapy.core.Direction.UP:
            actions = [
                momapy.drawing.MoveTo(self.joint3()),
                momapy.drawing.LineTo(self.joint1()),
                momapy.drawing.LineTo(self.joint2()),
            ]
        elif self.direction == momapy.core.Direction.LEFT:
            actions = [
                momapy.drawing.MoveTo(self.joint1()),
                momapy.drawing.LineTo(self.joint3()),
                momapy.drawing.LineTo(self.joint2()),
            ]
        elif self.direction == momapy.core.Direction.DOWN:
            actions = [
                momapy.drawing.MoveTo(self.joint1()),
                momapy.drawing.LineTo(self.joint3()),
                momapy.drawing.LineTo(self.joint2()),
            ]
        drawing_element = momapy.drawing.Path(actions=actions)
        return [drawing_element]


@dataclasses.dataclass(frozen=True, kw_only=True)
class To(momapy.core.Shape):
    """Class for to shapes"""

    position: momapy.geometry.Point
    width: float
    height: float
    direction: momapy.core.Direction

    def joint1(self):
        if (
            self.direction == momapy.core.Direction.RIGHT
            or self.direction == momapy.core.Direction.DOWN
        ):
            p = self.position + (-self.width / 2, -self.height / 2)
        elif self.direction == momapy.core.Direction.UP:
            p = self.position + (0, -self.height / 2)
        elif self.direction == momapy.core.Direction.LEFT:
            p = self.position + (self.width / 2, -self.height / 2)
        return p

    def joint2(self):
        if self.direction == momapy.core.Direction.RIGHT:
            p = self.position + (self.width / 2, 0)
        elif (
            self.direction == momapy.core.Direction.UP
            or self.direction == momapy.core.Direction.LEFT
        ):
            p = self.position + (self.width / 2, self.height / 2)
        elif self.direction == momapy.core.Direction.DOWN:
            p = self.position + (self.width / 2, -self.height / 2)
        return p

    def joint3(self):
        if (
            self.direction == momapy.core.Direction.RIGHT
            or self.direction == momapy.core.Direction.UP
        ):
            p = self.position + (-self.width / 2, self.height / 2)
        elif self.direction == momapy.core.Direction.LEFT:
            p = self.position + (-self.width / 2, 0)
        elif self.direction == momapy.core.Direction.DOWN:
            p = self.position + (0, self.height / 2)
        return p

    def drawing_elements(self):
        if self.direction == momapy.core.Direction.RIGHT:
            actions = [
                momapy.drawing.MoveTo(self.joint1()),
                momapy.drawing.EllipticalArc(
                    self.joint2(),
                    self.width,
                    self.height / 2,
                    0,
                    0,
                    0,
                ),
                momapy.drawing.EllipticalArc(
                    self.joint3(),
                    self.width,
                    self.height / 2,
                    0,
                    0,
                    0,
                ),
            ]
        elif self.direction == momapy.core.Direction.UP:
            actions = [
                momapy.drawing.MoveTo(self.joint3()),
                momapy.drawing.EllipticalArc(
                    self.joint1(),
                    self.height,
                    self.width / 2,
                    0,
                    0,
                    0,
                ),
                momapy.drawing.EllipticalArc(
                    self.joint2(),
                    self.height,
                    self.width / 2,
                    0,
                    0,
                    0,
                ),
            ]
        elif self.direction == momapy.core.Direction.LEFT:
            actions = [
                momapy.drawing.MoveTo(self.joint1()),
                momapy.drawing.EllipticalArc(
                    self.joint3(),
                    self.width,
                    self.height / 2,
                    0,
                    0,
                    1,
                ),
                momapy.drawing.EllipticalArc(
                    self.joint2(),
                    self.width,
                    self.height / 2,
                    0,
                    0,
                    1,
                ),
            ]
        elif self.direction == momapy.core.Direction.DOWN:
            actions = [
                momapy.drawing.MoveTo(self.joint1()),
                momapy.drawing.EllipticalArc(
                    self.joint3(),
                    self.height,
                    self.width / 2,
                    0,
                    0,
                    1,
                ),
                momapy.drawing.EllipticalArc(
                    self.joint2(),
                    self.height,
                    self.width / 2,
                    0,
                    0,
                    1,
                ),
            ]

        drawing_element = momapy.drawing.Path(actions=actions)
        return [drawing_element]



"""Classes and functions for building maps and map elements"""

import abc
import dataclasses
import typing
import typing_extensions
import types
import inspect

import frozendict

import momapy.monitoring


class Builder(abc.ABC, momapy.monitoring.Monitored):
    """Abstract class for builder objects"""

    _cls_to_build: typing.ClassVar[type]

    @abc.abstractmethod
    def build(
        self,
        inside_collections: bool = True,
        builder_to_object: dict[int, typing.Any] | None = None,
    ) -> typing.Any:
        """Build and return an object from the builder object"""
        pass

    @classmethod
    @abc.abstractmethod
    def from_object(
        cls,
        obj: typing.Any,
        inside_collections: bool = True,
        omit_keys: bool = True,
        object_to_builder: dict[int, "Builder"] | None = None,
    ) -> typing_extensions.Self:
        """Create and return a builder object from an object"""
        pass


builders = {}


def _transform_type(type_, make_optional=False, make_union=False):
    if isinstance(
        type_, typing.ForwardRef
    ):  # TO DO: should find if type is already in builders first
        new_type = typing.ForwardRef(f"{type_.__forward_arg__}Builder")
    else:
        # We get the origin of type_, e.g., if type_ = X[Y, Z, ...] we get X
        o_type = typing.get_origin(type_)  # returns None if not supported
        if o_type is not None:
            if isinstance(o_type, type):  # o_type is a type
                if o_type == types.UnionType:  # from t1 | t2 syntax
                    new_o_type = typing.Union
                else:
                    new_o_type = get_or_make_builder_cls(o_type)
            else:  # o_type is an object from typing
                new_o_type = o_type
            new_type = new_o_type[
                tuple([_transform_type(a_type) for a_type in typing.get_args(type_)])
            ]
        else:  # type_ has no origin
            if isinstance(type_, type):  # type_ is a type
                new_type = get_or_make_builder_cls(type_)
                if new_type is None:
                    new_type = type_
            else:
                new_type = type_
    if make_optional:
        new_type = typing.Optional[new_type]
    if make_union:
        new_type = typing.Union[type_, new_type]
    return new_type


def _make_builder_cls(
    cls, builder_fields=None, builder_bases=None, builder_namespace=None
):
    def _builder_build(
        self,
        inside_collections: bool = True,
        builder_to_object: dict[int, typing.Any] | None = None,
    ):
        if builder_to_object is not None:
            obj = builder_to_object.get(id(self))
            if obj is not None:
                return obj
        else:
            builder_to_object = {}
        args = {}
        for field in dataclasses.fields(self):
            attr_value = getattr(self, field.name)
            args[field.name] = object_from_builder(
                builder=attr_value,
                inside_collections=inside_collections,
                builder_to_object=builder_to_object,
            )
        obj = self._cls_to_build(**args)
        builder_to_object[id(self)] = obj
        return obj

    def _builder_from_object(
        cls,
        obj,
        inside_collections: bool = True,
        omit_keys: bool = True,
        object_to_builder: dict[int, "Builder"] | None = None,
    ):
        if object_to_builder is not None:
            builder = object_to_builder.get(id(obj))
            if builder is not None:
                return builder
        else:
            object_to_builder = {}
        args = {}
        for field_ in dataclasses.fields(obj):
            attr_value = getattr(obj, field_.name)
            args[field_.name] = builder_from_object(
                obj=attr_value,
                inside_collections=inside_collections,
                omit_keys=omit_keys,
                object_to_builder=object_to_builder,
            )
        builder = cls(**args)
        object_to_builder[id(obj)] = builder
        return builder

    if builder_fields is None:
        builder_fields = []
    if builder_bases is None:
        builder_bases = []
    if builder_namespace is None:
        builder_namespace = {}
    # We transform the fields to builder fields
    cls_fields = dataclasses.fields(cls)
    builder_field_names = set([builder_field[0] for builder_field in builder_fields])
    for field_ in cls_fields:
        field_name = field_.name
        # We only consider fields that are not already in the input fields
        if field_name not in builder_field_names:
            field_dict = {}
            has_default = False
            if field_.default_factory != dataclasses.MISSING:
                if isinstance(field_.default_factory, type):
                    field_dict["default_factory"] = _transform_type(
                        field_.default_factory
                    )
                else:  # in case of a func for example
                    field_dict["default_factory"] = field_.default_factory
                has_default = True
            if field_.default != dataclasses.MISSING:
                field_dict["default"] = field_.default  # TO DO: transform?
                has_default = True
            if not has_default:
                field_dict["default"] = None
            field_type = _transform_type(
                field_.type, make_optional=not has_default, make_union=True
            )
            builder_fields.append(
                (field_name, field_type, dataclasses.field(**field_dict))
            )
    builder_namespace["build"] = _builder_build
    builder_namespace["from_object"] = classmethod(_builder_from_object)
    builder_namespace["_cls_to_build"] = cls
    # We add the undundered methods from the non-builder class
    # Do we really want this? Should we keep builders really only to build?
    for member in inspect.getmembers(cls):
        func_name = member[0]
        func = member[1]

        if not func_name.startswith("__") and not func_name == "_cls_to_build":
            builder_namespace[func_name] = func
    # We add the transformed bases
    cls_bases = [get_or_make_builder_cls(base_cls) for base_cls in cls.__bases__]
    builder_bases = builder_bases + [
        base_cls for base_cls in cls_bases if issubclass(base_cls, Builder)
    ]
    # We add the Builder class to the bases
    has_builder_cls = False
    for builder_base in builder_bases:
        if Builder in builder_base.__mro__:
            has_builder_cls = True
            break
    if not has_builder_cls:
        builder_bases = [Builder] + builder_bases
    builder_bases = tuple(builder_bases)

    builder = dataclasses.make_dataclass(
        cls_name=f"{cls.__name__}Builder",
        fields=builder_fields,
        bases=builder_bases,
        namespace=builder_namespace,
        eq=False,
        kw_only=False,
    )
    return builder


def object_from_builder(
    builder: Builder,
    inside_collections=True,
    builder_to_object: dict[int, typing.Any] | None = None,
):
    """Create and return an object from a builder object"""
    if builder_to_object is not None:
        if id(builder) in builder_to_object:
            return builder_to_object[id(builder)]
    else:
        builder_to_object = {}
    if isinstance(builder, Builder):
        obj = builder.build(
            inside_collections=inside_collections,
            builder_to_object=builder_to_object,
        )
        builder_to_object[id(builder)] = obj
        return obj
    if inside_collections:
        if isinstance(builder, (list, tuple, set, frozenset)):
            return type(builder)(
                [
                    object_from_builder(
                        builder=e,
                        inside_collections=inside_collections,
                        builder_to_object=builder_to_object,
                    )
                    for e in builder
                ]
            )
        elif isinstance(builder, (dict, frozendict.frozendict)):
            return type(builder)(
                [
                    (
                        object_from_builder(
                            builder=k,
                            inside_collections=inside_collections,
                            builder_to_object=builder_to_object,
                        ),
                        object_from_builder(
                            builder=v,
                            inside_collections=inside_collections,
                            builder_to_object=builder_to_object,
                        ),
                    )
                    for k, v in builder.items()
                ]
            )
    return builder


def builder_from_object(
    obj: typing.Any,
    inside_collections=True,
    omit_keys=True,
    object_to_builder: dict[int, "Builder"] | None = None,
) -> Builder:
    """Create and return a builder object from an object"""
    if object_to_builder is not None:
        builder = object_to_builder.get(id(obj))
        if builder is not None:
            return builder
    else:
        object_to_builder = {}
    cls = get_or_make_builder_cls(type(obj))
    if issubclass(cls, Builder):
        return cls.from_object(
            obj=obj,
            inside_collections=inside_collections,
            omit_keys=omit_keys,
            object_to_builder=object_to_builder,
        )
    if inside_collections:
        if isinstance(obj, (list, tuple, set, frozenset)):
            return type(obj)(
                [
                    builder_from_object(
                        obj=e,
                        inside_collections=inside_collections,
                        omit_keys=omit_keys,
                        object_to_builder=object_to_builder,
                    )
                    for e in obj
                ]
            )
        elif isinstance(obj, (dict, frozendict.frozendict)):
            return type(obj)(
                [
                    (
                        (
                            builder_from_object(
                                obj=k,
                                inside_collections=inside_collections,
                                omit_keys=omit_keys,
                                object_to_builder=object_to_builder,
                            ),
                            builder_from_object(
                                obj=v,
                                inside_collections=inside_collections,
                                omit_keys=omit_keys,
                                object_to_builder=object_to_builder,
                            ),
                        )
                        if not omit_keys
                        else (
                            k,
                            builder_from_object(
                                obj=v,
                                inside_collections=inside_collections,
                                omit_keys=omit_keys,
                                object_to_builder=object_to_builder,
                            ),
                        )
                    )
                    for k, v in obj.items()
                ]
            )
    return obj


def new_builder_object(cls: typing.Type, *args, **kwargs) -> Builder:
    """Create and return a builder object from an object class or a builder class"""
    if not issubclass(cls, Builder):
        cls = get_or_make_builder_cls(cls)
    return cls(*args, **kwargs)


def get_or_make_builder_cls(
    cls: typing.Type,
    builder_fields: (
        typing.Collection[tuple[str, typing.Type, dataclasses.Field]] | None
    ) = None,
    builder_bases: typing.Collection[typing.Type] | None = None,
    builder_namespace: dict[str, typing.Any] | None = None,
) -> typing.Type:
    """Get and return an existing builder class for the given class or make and return a new builder class for it"""
    builder_cls = get_builder_cls(cls)
    if builder_cls is None:
        if dataclasses.is_dataclass(cls):
            builder_cls = _make_builder_cls(
                cls, builder_fields, builder_bases, builder_namespace
            )
            register_builder_cls(builder_cls)
        else:
            builder_cls = cls
    return builder_cls


def has_builder_cls(cls: typing.Type) -> bool:
    """Return `true` if there is a registered builder class for the given class, and `false` otherwise"""
    return cls in builders


def get_builder_cls(cls: typing.Type) -> typing.Type:
    """Return the builder class registered for the given class or `None` if no builder class is registered for that class"""
    return builders.get(cls)


def register_builder_cls(builder_cls: typing.Type) -> None:
    """Register a builder class"""
    builders[builder_cls._cls_to_build] = builder_cls


def isinstance_or_builder(
    obj: typing.Any, type_: typing.Type | tuple[typing.Type]
) -> bool:
    """Return `true` if the object is an istance of the given classes or of their registered builder classes, and `false` otherwise"""
    if isinstance(type_, type):
        type_ = (type_,)
    type_ += tuple([get_or_make_builder_cls(t) for t in type_])
    return isinstance(obj, type_)


def issubclass_or_builder(
    cls: typing.Type, type_: typing.Type | tuple[typing.Type]
) -> bool:
    """Return `true` if the class is a subclass of the given classes or of their registered builder classes, and `false` otherwise"""
    if isinstance(type_, type):
        type_ = (type_,)
    type_ += tuple([get_or_make_builder_cls(t) for t in type_])
    return issubclass(cls, type_)


def super_or_builder(type_: typing.Type, obj: typing.Any) -> typing.Type:
    """Return the super class for a given class or its builder class and an object"""
    try:
        s = super(type_, obj)
    except TypeError:
        builder = get_or_make_builder_cls(type_)
        s = super(builder, obj)
    finally:
        return s



import collections

import momapy.positioning
import momapy.builder
import momapy.sbgn.pd
import momapy.sbgn.af


def set_compartments_to_fit_content(map_, xsep=0, ysep=0):
    if isinstance(map_, momapy.sbgn.core.SBGNMap):
        map_builder = momapy.builder.builder_from_object(map_)
    else:
        map_builder = map_
    compartment_entities_mapping = collections.defaultdict(list)
    model = map_builder.model
    if momapy.builder.isinstance_or_builder(map_builder, momapy.sbgn.pd.SBGNPDMap):
        for entity_pool in model.entity_pools:
            compartment = entity_pool.compartment
            if compartment is not None:
                compartment_entities_mapping[compartment].append(entity_pool)
    else:
        for activity in model.activities:
            compartment = activity.compartment
            if compartment is not None:
                compartment_entities_mapping[compartment].append(activity)
    for compartment in compartment_entities_mapping:
        for compartment_layout in map_builder.get_mapping(compartment):
            elements = []
            for entity in compartment_entities_mapping[compartment]:
                for entity_layout in map_builder.get_mapping(entity):
                    elements.append(entity_layout)
            momapy.positioning.set_fit(compartment_layout, elements, xsep, ysep)
            if compartment_layout.label is not None:
                compartment_layout.label.position = compartment_layout.position
                compartment_layout.label.width = compartment_layout.width
                compartment_layout.label.height = compartment_layout.height
    if isinstance(map_, momapy.sbgn.core.SBGNMap):
        return momapy.builder.object_from_builder(map_builder)
    return map_builder


def set_complexes_to_fit_content(map_, xsep=0, ysep=0):
    if isinstance(map_, momapy.sbgn.core.SBGNMap):
        map_builder = momapy.builder.builder_from_object(map_)
    else:
        map_builder = map_
    for entity_pool in map_builder.model.entity_pools:
        if isinstance(
            entity_pool,
            momapy.builder.get_or_make_builder_cls(momapy.sbgn.pd.Complex),
        ):
            for complex_layout in map_builder.get_mapping(entity_pool):
                elements = []
                for subunit in entity_pool.subunits:
                    subunit_layouts = map_builder.get_mapping((subunit, entity_pool))
                    for subunit_layout in subunit_layouts:
                        if subunit_layout in complex_layout.layout_elements:
                            elements.append(subunit_layout)
                if len(elements) > 0:
                    momapy.positioning.set_fit(complex_layout, elements, xsep, ysep)
                    if complex_layout.label is not None:
                        complex_layout.label.position = complex_layout.position
                        complex_layout.label.width = complex_layout.width
                        complex_layout.label.height = complex_layout.height
    if isinstance(map_, momapy.sbgn.core.SBGNMap):
        return momapy.builder.object_from_builder(map_builder)
    return map_builder


def set_submaps_to_fit_content(map_, xsep=0, ysep=0):
    if isinstance(map_, momapy.sbgn.core.SBGNMap):
        map_builder = momapy.builder.builder_from_object(map_)
    else:
        map_builder = map_
    for submap in map_builder.model.submaps:
        for submap_layout in map_builder.get_mapping(submap):
            elements = []
            for terminal in submap.terminals:
                terminal_layouts = map_builder.layout_model_mapping.get_mapping(
                    (
                        terminal,
                        submap,
                    )
                )
                for terminal_layout in terminal_layouts:
                    if terminal_layout in submap_layout.layout_elements:
                        elements.append(terminal_layout)
            if len(elements) > 0:
                momapy.positioning.set_fit(submap_layout, elements, xsep, ysep)
                if submap_layout.label is not None:
                    submap_layout.label.position = submap_layout.position
                    submap_layout.label.width = submap_layout.width
                    submap_layout.label.height = submap_layout.height
    if isinstance(map_, momapy.sbgn.core.SBGNMap):
        return momapy.builder.object_from_builder(map_builder)
    return map_builder


def set_nodes_to_fit_labels(
    map_,
    xsep=0,
    ysep=0,
    omit_width=False,
    omit_height=False,
    restrict_to=None,
    exclude=None,
):
    if isinstance(map_, momapy.sbgn.core.SBGNMap):
        map_builder = momapy.builder.builder_from_object(map_)
    else:
        map_builder = map_
    if restrict_to is None:
        restrict_to = []
    if not restrict_to:
        restrict_to = [momapy.core.Node]
    if exclude is None:
        exclude = []
    exclude = tuple(exclude)
    restrict_to = tuple(restrict_to)
    if omit_width and omit_height:
        return
    for layout_element in map_builder.layout.descendants():
        if (
            momapy.builder.isinstance_or_builder(layout_element, restrict_to)
            and not momapy.builder.isinstance_or_builder(layout_element, exclude)
            and hasattr(layout_element, "label")
            and layout_element.label is not None
        ):
            bbox = momapy.positioning.fit([layout_element.label.bbox()], xsep, ysep)
            if not omit_width:
                if bbox.width > layout_element.width:
                    layout_element.width = bbox.width
            if not omit_height:
                if bbox.height > layout_element.height:
                    layout_element.height = bbox.height
            momapy.positioning.set_position(
                layout_element, bbox.position, anchor="label_center"
            )
    if isinstance(map_, momapy.sbgn.core.SBGNMap):
        return momapy.builder.object_from_builder(map_builder)
    return map_builder


def set_arcs_to_borders(map_):
    def _set_arc_to_borders(
        arc_layout_element, source, source_type, target, target_type
    ):
        points = arc_layout_element.points()
        if source_type == "left":
            start_point = source.left_connector_tip()
        elif source_type == "right":
            start_point = source.right_connector_tip()
        else:
            if len(arc_layout_element.segments) > 1:
                start_reference_point = points[1]
            else:
                if target_type == "border":
                    start_reference_point = target.center()
                elif target_type == "left":
                    start_reference_point = target.left_connector_tip()
                else:
                    start_reference_point = target.right_connector_tip()
            start_point = source.border(start_reference_point)
        if target_type == "left":
            end_point = target.left_connector_tip()
        elif target_type == "right":
            end_point = target.right_connector_tip()
        else:
            if len(arc_layout_element.segments) > 1:
                end_reference_point = points[-2]
            else:
                if source_type == "border":
                    end_reference_point = source.center()
                elif source_type == "left":
                    end_reference_point = source.left_connector_tip()
                else:
                    end_reference_point = source.right_connector_tip()
            end_point = target.border(end_reference_point)
        arc_layout_element.segments[0].p1 = momapy.builder.builder_from_object(
            start_point
        )
        arc_layout_element.segments[-1].p2 = momapy.builder.builder_from_object(
            end_point
        )

    if isinstance(map_, momapy.sbgn.core.SBGNMap):
        map_builder = momapy.builder.builder_from_object(map_)
    else:
        map_builder = map_
    for layout_element in map_builder.layout.layout_elements:
        # Flux arcs
        if momapy.builder.isinstance_or_builder(
            layout_element,
            (
                momapy.sbgn.pd.GenericProcessLayout,
                momapy.sbgn.pd.AssociationLayout,
                momapy.sbgn.pd.DissociationLayout,
                momapy.sbgn.pd.OmittedProcessLayout,
                momapy.sbgn.pd.UncertainProcessLayout,
            ),
        ):
            for sub_layout_element in layout_element.layout_elements:
                if momapy.builder.isinstance_or_builder(
                    sub_layout_element, (momapy.sbgn.pd.ConsumptionLayout)
                ):
                    if layout_element.left_to_right:
                        source_type = "left"
                    else:
                        source_type = "right"
                    _set_arc_to_borders(
                        sub_layout_element,
                        layout_element,
                        source_type,
                        sub_layout_element.target,
                        "border",
                    )
                elif momapy.builder.isinstance_or_builder(
                    sub_layout_element, (momapy.sbgn.pd.ProductionLayout)
                ):
                    product, _ = map_builder.get_mapping(sub_layout_element)
                    if momapy.builder.isinstance_or_builder(
                        product, momapy.sbgn.pd.Product
                    ):
                        if layout_element.left_to_right:
                            source_type = "right"
                        else:
                            source_type = "left"
                    else:
                        if layout_element.left_to_right:
                            source_type = "left"
                        else:
                            source_type = "right"
                    _set_arc_to_borders(
                        sub_layout_element,
                        layout_element,
                        source_type,
                        sub_layout_element.target,
                        "border",
                    )
        # Logical arcs
        elif momapy.builder.isinstance_or_builder(
            layout_element,
            (
                momapy.sbgn.pd.AndOperatorLayout,
                momapy.sbgn.pd.OrOperatorLayout,
                momapy.sbgn.pd.NotOperatorLayout,
                momapy.sbgn.af.AndOperatorLayout,
                momapy.sbgn.af.OrOperatorLayout,
                momapy.sbgn.af.NotOperatorLayout,
                momapy.sbgn.af.DelayOperatorLayout,
                momapy.sbgn.pd.EquivalenceOperatorLayout,
            ),
        ):
            for sub_layout_element in layout_element.layout_elements:
                if momapy.builder.isinstance_or_builder(
                    sub_layout_element,
                    (
                        momapy.sbgn.pd.LogicArcLayout,
                        momapy.sbgn.af.LogicArcLayout,
                    ),
                ):
                    target = sub_layout_element.target
                    if layout_element.left_to_right:
                        source_type = "left"
                    else:
                        source_type = "right"
                if momapy.builder.isinstance_or_builder(
                    target,
                    (
                        momapy.sbgn.pd.AndOperatorLayout,
                        momapy.sbgn.pd.OrOperatorLayout,
                        momapy.sbgn.pd.NotOperatorLayout,
                        momapy.sbgn.af.AndOperatorLayout,
                        momapy.sbgn.af.OrOperatorLayout,
                        momapy.sbgn.af.NotOperatorLayout,
                        momapy.sbgn.af.DelayOperatorLayout,
                        momapy.sbgn.pd.EquivalenceOperatorLayout,
                    ),
                ):
                    if target.left_to_right:
                        target_type = "left"
                    else:
                        target_type = "right"
                else:
                    target_type = "border"
                _set_arc_to_borders(
                    sub_layout_element,
                    layout_element,
                    source_type,
                    target,
                    target_type,
                )
        elif momapy.builder.isinstance_or_builder(
            layout_element,
            (
                momapy.sbgn.pd.ModulationLayout,
                momapy.sbgn.pd.StimulationLayout,
                momapy.sbgn.pd.CatalysisLayout,
                momapy.sbgn.pd.NecessaryStimulationLayout,
                momapy.sbgn.pd.InhibitionLayout,
                momapy.sbgn.af.UnknownInfluenceLayout,
                momapy.sbgn.af.PositiveInfluenceLayout,
                momapy.sbgn.af.NecessaryStimulationLayout,
                momapy.sbgn.af.NegativeInfluenceLayout,
            ),
        ):
            source = layout_element.source
            if momapy.builder.isinstance_or_builder(
                source,
                (
                    momapy.sbgn.pd.AndOperatorLayout,
                    momapy.sbgn.pd.OrOperatorLayout,
                    momapy.sbgn.pd.NotOperatorLayout,
                    momapy.sbgn.af.AndOperatorLayout,
                    momapy.sbgn.af.OrOperatorLayout,
                    momapy.sbgn.af.NotOperatorLayout,
                    momapy.sbgn.af.DelayOperatorLayout,
                ),
            ):
                if source.left_to_right:
                    source_type = "right"
                else:
                    source_type = "left"
            else:
                source_type = "border"
            _set_arc_to_borders(
                layout_element,
                source,
                source_type,
                layout_element.target,
                "border",
            )
    if isinstance(map_, momapy.sbgn.core.SBGNMap):
        return momapy.builder.object_from_builder(map_builder)
    return map_builder


def set_auxiliary_units_to_borders(map_):
    def _rec_set_auxiliary_units_to_borders(layout_element):
        for child in layout_element.children():
            if momapy.builder.isinstance_or_builder(
                child,
                (
                    momapy.sbgn.pd.StateVariableLayout,
                    momapy.sbgn.pd.UnitOfInformationLayout,
                    momapy.sbgn.af.UnspecifiedEntityUnitOfInformationLayout,
                    momapy.sbgn.af.MacromoleculeUnitOfInformationLayout,
                    momapy.sbgn.af.NucleicAcidFeatureUnitOfInformationLayout,
                    momapy.sbgn.af.ComplexUnitOfInformationLayout,
                    momapy.sbgn.af.SimpleChemicalUnitOfInformationLayout,
                    momapy.sbgn.af.PerturbationUnitOfInformationLayout,
                ),
            ):
                position = layout_element.self_border(child.position)
                child.position = position
                if child.label is not None:
                    child.label.position = position
            _rec_set_auxiliary_units_to_borders(child)

    if isinstance(map_, momapy.sbgn.core.SBGNMap):
        map_builder = momapy.builder.builder_from_object(map_)
    else:
        map_builder = map_
    _rec_set_auxiliary_units_to_borders(map_builder.layout)
    if isinstance(map_, momapy.sbgn.core.SBGNMap):
        return momapy.builder.object_from_builder(map_builder)
    return map_builder


def set_auxiliary_units_label_font_size(map_, font_size: float):
    def _rec_set_auxiliary_units_label_font_size(layout_element, font_size: float):
        for child in layout_element.children():
            if momapy.builder.isinstance_or_builder(
                child,
                (
                    momapy.sbgn.pd.StateVariableLayout,
                    momapy.sbgn.pd.UnitOfInformationLayout,
                    momapy.sbgn.af.UnspecifiedEntityUnitOfInformationLayout,
                    momapy.sbgn.af.MacromoleculeUnitOfInformationLayout,
                    momapy.sbgn.af.NucleicAcidFeatureUnitOfInformationLayout,
                    momapy.sbgn.af.ComplexUnitOfInformationLayout,
                    momapy.sbgn.af.SimpleChemicalUnitOfInformationLayout,
                    momapy.sbgn.af.PerturbationUnitOfInformationLayout,
                ),
            ):
                if child.label is not None:
                    child.label.font_size = font_size
            _rec_set_auxiliary_units_label_font_size(child, font_size)

    if isinstance(map_, momapy.sbgn.core.SBGNMap):
        map_builder = momapy.builder.builder_from_object(map_)
    else:
        map_builder = map_
    _rec_set_auxiliary_units_label_font_size(map_builder.layout, font_size)
    if isinstance(map_, momapy.sbgn.core.SBGNMap):
        return momapy.builder.object_from_builder(map_builder)
    return map_builder


def set_layout_to_fit_content(map_, xsep=0, ysep=0):
    if isinstance(map_, momapy.sbgn.core.SBGNMap):
        map_builder = momapy.builder.builder_from_object(map_)
    else:
        map_builder = map_
    momapy.positioning.set_fit(
        map_builder.layout, map_builder.layout.layout_elements, xsep, ysep
    )
    if isinstance(map_, momapy.sbgn.core.SBGNMap):
        return momapy.builder.object_from_builder(map_builder)
    return map_builder


def tidy(
    map_,
    auxiliary_units_omit_width=False,
    auxiliary_units_omit_height=False,
    nodes_xsep=0,
    nodes_ysep=0,
    auxiliary_units_xsep=0,
    auxiliary_units_ysep=0,
    complexes_xsep=0,
    complexes_ysep=0,
    compartments_xsep=0,
    compartments_ysep=0,
    layout_xsep=0,
    layout_ysep=0,
):
    if isinstance(map_, momapy.sbgn.core.SBGNMap):
        map_builder = momapy.builder.builder_from_object(map_)
    else:
        map_builder = map_
    set_nodes_to_fit_labels(
        map_builder,
        xsep=nodes_xsep,
        ysep=nodes_ysep,
        exclude=[
            momapy.sbgn.pd.StateVariableLayout,
            momapy.sbgn.pd.UnitOfInformationLayout,
            momapy.sbgn.pd.ComplexLayout,
        ],
    )
    set_auxiliary_units_to_borders(map_builder)
    set_nodes_to_fit_labels(
        map_builder,
        xsep=auxiliary_units_xsep,
        ysep=auxiliary_units_ysep,
        omit_width=auxiliary_units_omit_width,
        omit_height=auxiliary_units_omit_height,
        restrict_to=[
            momapy.sbgn.pd.StateVariableLayout,
            momapy.sbgn.pd.UnitOfInformationLayout,
        ],
    )
    if momapy.builder.isinstance_or_builder(map_builder, momapy.sbgn.pd.SBGNPDMap):
        set_complexes_to_fit_content(map_builder, complexes_xsep, complexes_ysep)
    set_submaps_to_fit_content(map_builder, 0, 0)
    set_compartments_to_fit_content(map_builder, compartments_xsep, compartments_ysep)
    set_arcs_to_borders(map_builder)
    set_layout_to_fit_content(map_builder, layout_xsep, layout_ysep)
    if isinstance(map_, momapy.sbgn.core.SBGNMap):
        return momapy.builder.object_from_builder(map_builder)
    return map_builder


def sbgned_tidy(map_):
    return tidy(
        map_,
        auxiliary_units_omit_width=False,
        auxiliary_units_omit_height=True,
        nodes_xsep=0,
        nodes_ysep=0,
        auxiliary_units_xsep=10,
        auxiliary_units_ysep=0,
        complexes_xsep=5,
        complexes_ysep=5,
        compartments_xsep=10,
        compartments_ysep=10,
        layout_xsep=0,
        layout_ysep=0,
    )


def newt_tidy(map_):
    return tidy(
        map_,
        auxiliary_units_omit_width=False,
        auxiliary_units_omit_height=True,
        nodes_xsep=0,
        nodes_ysep=0,
        auxiliary_units_xsep=1,
        auxiliary_units_ysep=0,
        complexes_xsep=10,
        complexes_ysep=10,
        compartments_xsep=14,
        compartments_ysep=14,
        layout_xsep=0,
        layout_ysep=0,
    )



"""Base classes defining core SBGN map elements"""

import abc
import dataclasses
import typing

import momapy.core
import momapy.geometry


@dataclasses.dataclass(frozen=True, kw_only=True)
class SBGNModelElement(momapy.core.ModelElement):
    """Base class for SBGN model elements"""

    pass


@dataclasses.dataclass(frozen=True, kw_only=True)
class SBGNAuxiliaryUnit(SBGNModelElement):
    """Base class for SBGN auxiliary units"""

    pass


@dataclasses.dataclass(frozen=True, kw_only=True)
class SBGNRole(SBGNModelElement):
    """Base class for SBGN roles"""

    element: SBGNModelElement


@dataclasses.dataclass(frozen=True, kw_only=True)
class SBGNModel(momapy.core.Model):
    """Base class for SBGN models"""


@dataclasses.dataclass(frozen=True, kw_only=True)
class SBGNLayout(momapy.core.Layout):
    """Base class for SBGN layouts"""

    fill: momapy.drawing.NoneValueType | momapy.coloring.Color | None = (
        momapy.coloring.white
    )


@dataclasses.dataclass(frozen=True, kw_only=True)
class SBGNMap(momapy.core.Map):
    """Base class for SBGN maps"""

    model: SBGNModel
    layout: SBGNLayout


@dataclasses.dataclass(frozen=True)
class SBGNNode(momapy.core.Node):
    """Base class for SBGN nodes"""

    fill: momapy.drawing.NoneValueType | momapy.coloring.Color | None = (
        momapy.coloring.white
    )
    stroke: momapy.drawing.NoneValueType | momapy.coloring.Color | None = (
        momapy.coloring.black
    )
    stroke_width: float | None = 1.25

    def _border_drawing_elements(self):
        drawing_elements = []
        done_bases = []
        for base in type(self).__mro__:
            if (
                momapy.builder.issubclass_or_builder(base, _SBGNMixin)
                and base is not type(self)
                and not any(
                    [issubclass(done_base, base) for done_base in done_bases]
                )
            ):
                drawing_elements += getattr(base, "_mixin_drawing_elements")(
                    self
                )
                done_bases.append(base)
        return drawing_elements


@dataclasses.dataclass(frozen=True)
class SBGNSingleHeadedArc(momapy.core.SingleHeadedArc):
    """Base class for SBGN single-headed arcs"""

    arrowhead_fill: (
        momapy.drawing.NoneValueType | momapy.coloring.Color | None
    ) = momapy.coloring.white
    arrowhead_stroke: (
        momapy.drawing.NoneValueType | momapy.coloring.Color | None
    ) = momapy.coloring.black
    arrowhead_stroke_width: float | None = 1.25
    path_fill: momapy.drawing.NoneValueType | momapy.coloring.Color | None = (
        momapy.drawing.NoneValue
    )
    path_stroke: (
        momapy.drawing.NoneValueType | momapy.coloring.Color | None
    ) = momapy.coloring.black
    path_stroke_width: float | None = 1.25


@dataclasses.dataclass(frozen=True)
class SBGNDoubleHeadedArc(momapy.core.DoubleHeadedArc):
    """Base class for SBGN double-headed arcs"""

    end_arrowhead_fill: (
        momapy.drawing.NoneValueType | momapy.coloring.Color | None
    ) = momapy.coloring.white
    end_arrowhead_stroke: (
        momapy.drawing.NoneValueType | momapy.coloring.Color | None
    ) = momapy.coloring.black
    end_arrowhead_stroke_width: float | None = 1.25
    path_fill: momapy.drawing.NoneValueType | momapy.coloring.Color | None = (
        momapy.drawing.NoneValue
    )
    path_stroke: (
        momapy.drawing.NoneValueType | momapy.coloring.Color | None
    ) = momapy.coloring.black
    path_stroke_width: float | None = 1.25
    start_arrowhead_fill: (
        momapy.drawing.NoneValueType | momapy.coloring.Color | None
    ) = momapy.coloring.white
    start_arrowhead_stroke: (
        momapy.drawing.NoneValueType | momapy.coloring.Color | None
    ) = momapy.coloring.black
    start_arrowhead_stroke_width: float | None = 1.25


@dataclasses.dataclass(frozen=True)
class _SBGNMixin(object):
    @classmethod
    @abc.abstractmethod
    def _mixin_drawing_elements(cls, obj):
        pass


@dataclasses.dataclass(frozen=True, kw_only=True)
class _ConnectorsMixin(_SBGNMixin):
    direction: momapy.core.Direction = momapy.core.Direction.HORIZONTAL
    left_to_right: bool = True
    left_connector_length: float = 10.0
    right_connector_length: float = 10.0
    left_connector_stroke: (
        momapy.drawing.NoneValueType | momapy.coloring.Color | None
    ) = None
    left_connector_stroke_width: float | None = None
    left_connector_stroke_dasharray: (
        momapy.drawing.NoneValueType | tuple[float] | None
    ) = None
    left_connector_stroke_dashoffset: float | None = None
    left_connector_fill: (
        momapy.drawing.NoneValueType | momapy.coloring.Color | None
    ) = None
    left_connector_transform: (
        momapy.drawing.NoneValueType
        | tuple[momapy.geometry.Transformation]
        | None
    ) = None
    left_connector_filter: (
        momapy.drawing.NoneValueType | momapy.drawing.Filter | None
    ) = None
    right_connector_stroke: (
        momapy.drawing.NoneValueType | momapy.coloring.Color | None
    ) = None
    right_connector_stroke_width: float | None = None
    right_connector_stroke_dasharray: (
        momapy.drawing.NoneValueType | tuple[float] | None
    ) = None
    right_connector_stroke_dashoffset: float | None = None
    right_connector_fill: (
        momapy.drawing.NoneValueType | momapy.coloring.Color | None
    ) = None
    right_connector_transform: (
        momapy.drawing.NoneValueType
        | tuple[momapy.geometry.Transformation]
        | None
    ) = None
    right_connector_filter: (
        momapy.drawing.NoneValueType | momapy.drawing.Filter | None
    ) = None

    def left_connector_base(self):
        if self.direction == momapy.core.Direction.VERTICAL:
            return momapy.geometry.Point(self.x, self.y - self.height / 2)
        else:
            return momapy.geometry.Point(self.x - self.width / 2, self.y)

    def right_connector_base(self):
        if self.direction == momapy.core.Direction.VERTICAL:
            return momapy.geometry.Point(self.x, self.y + self.height / 2)
        else:
            return momapy.geometry.Point(self.x + self.width / 2, self.y)

    def left_connector_tip(self):
        if self.direction == momapy.core.Direction.VERTICAL:
            return momapy.geometry.Point(
                self.x, self.y - self.height / 2 - self.left_connector_length
            )
        else:
            return momapy.geometry.Point(
                self.x - self.width / 2 - self.left_connector_length, self.y
            )

    def right_connector_tip(self):
        if self.direction == momapy.core.Direction.VERTICAL:
            return momapy.geometry.Point(
                self.x, self.y + self.height / 2 + self.right_connector_length
            )
        else:
            return momapy.geometry.Point(
                self.x + self.width / 2 + self.right_connector_length, self.y
            )

    def west(self):
        if self.direction == momapy.core.Direction.VERTICAL:
            return momapy.geometry.Point(self.x - self.width / 2, self.y)
        else:
            return momapy.geometry.Point(
                self.x - self.width / 2 - self.left_connector_length, self.y
            )

    def south(self):
        if self.direction == momapy.core.Direction.VERTICAL:
            return momapy.geometry.Point(
                self.x, self.y + self.height / 2 + self.right_connector_length
            )
        else:
            return momapy.geometry.Point(self.x, self.y + self.height / 2)

    def east(self):
        if self.direction == momapy.core.Direction.VERTICAL:
            return momapy.geometry.Point(self.x + self.width / 2, self.y)
        else:
            return momapy.geometry.Point(
                self.x + self.width / 2 + self.right_connector_length, self.y
            )

    def north(self):
        if self.direction == momapy.core.Direction.VERTICAL:
            return momapy.geometry.Point(
                self.x, self.y - self.height / 2 - self.left_connector_length
            )
        else:
            return momapy.geometry.Point(self.x, self.y - self.height / 2)

    @classmethod
    def _mixin_drawing_elements(cls, obj):
        if obj.direction == momapy.core.Direction.VERTICAL:
            left_actions = [
                momapy.drawing.MoveTo(obj.left_connector_base()),
                momapy.drawing.LineTo(
                    obj.left_connector_base() - (0, obj.left_connector_length)
                ),
            ]
            right_actions = [
                momapy.drawing.MoveTo(obj.right_connector_base()),
                momapy.drawing.LineTo(
                    obj.right_connector_base()
                    + (0, obj.right_connector_length)
                ),
            ]
        else:
            left_actions = [
                momapy.drawing.MoveTo(obj.left_connector_base()),
                momapy.drawing.LineTo(
                    obj.left_connector_base() - (obj.left_connector_length, 0)
                ),
            ]
            right_actions = [
                momapy.drawing.MoveTo(obj.right_connector_base()),
                momapy.drawing.LineTo(
                    obj.right_connector_base()
                    + (obj.right_connector_length, 0)
                ),
            ]
        path_left = momapy.drawing.Path(
            stroke=obj.left_connector_stroke,
            stroke_width=obj.left_connector_stroke_width,
            stroke_dasharray=obj.left_connector_stroke_dasharray,
            stroke_dashoffset=obj.left_connector_stroke_dashoffset,
            fill=obj.left_connector_fill,
            transform=obj.left_connector_transform,
            filter=obj.left_connector_filter,
            actions=left_actions,
        )
        path_right = momapy.drawing.Path(
            stroke=obj.right_connector_stroke,
            stroke_width=obj.right_connector_stroke_width,
            stroke_dasharray=obj.right_connector_stroke_dasharray,
            stroke_dashoffset=obj.right_connector_stroke_dashoffset,
            fill=obj.right_connector_fill,
            transform=obj.right_connector_transform,
            filter=obj.right_connector_filter,
            actions=right_actions,
        )
        return [path_left, path_right]


@dataclasses.dataclass(frozen=True)
class _SimpleMixin(_SBGNMixin):
    @abc.abstractmethod
    def _make_shape(self):
        pass

    @classmethod
    def _mixin_drawing_elements(cls, obj):
        shape = obj._make_shape()
        drawing_elements = shape.drawing_elements()
        return drawing_elements


@dataclasses.dataclass(frozen=True)
class _MultiMixin(_SBGNMixin):
    _n: typing.ClassVar[int] = 2
    offset: float = 3.0
    subunits_stroke: (
        tuple[momapy.drawing.NoneValueType | momapy.coloring.Color] | None
    ) = None
    subunits_stroke_width: (
        tuple[momapy.drawing.NoneValueType | float] | None
    ) = None
    subunits_stroke_dasharray: (
        tuple[momapy.drawing.NoneValueType | tuple[float]] | None
    ) = None
    subunits_stroke_dashoffset: tuple[float] | None = None
    subunits_fill: (
        tuple[momapy.drawing.NoneValueType | momapy.coloring.Color] | None
    ) = None
    subunits_transform: (
        tuple[
            momapy.drawing.NoneValueType
            | tuple[momapy.geometry.Transformation]
        ]
        | None
    ) = None
    subunits_filter: (
        tuple[momapy.drawing.NoneValueType | momapy.drawing.Filter] | None
    ) = None

    def _make_subunit_shape(
        self,
        position,
        width,
        height,
    ):
        pass

    @classmethod
    def _mixin_drawing_elements(cls, obj):
        drawing_elements = []
        width = obj.width - obj.offset * (obj._n - 1)
        height = obj.height - obj.offset * (obj._n - 1)
        for i in range(obj._n):
            position = obj.position + (
                obj.width / 2 - width / 2 - i * obj.offset,
                obj.height / 2 - height / 2 - i * obj.offset,
            )
            kwargs = {}
            for attr_name in [
                "stroke",
                "stroke_width",
                "stroke_dasharray",
                "stroke_dashoffset",
                "fill",
                "transform",
                "filter",
            ]:
                attr_value = getattr(obj, f"subunits_{attr_name}")
                if attr_value is not None and len(attr_value) > i:
                    kwargs[f"{attr_name}"] = attr_value[i]
            subunit_shape = obj._make_subunit_shape(position, width, height)
            kwargs["elements"] = subunit_shape.drawing_elements()
            group = momapy.drawing.Group(**kwargs)
            drawing_elements.append(group)
        return drawing_elements

    def label_center(self):
        width = self.width - self.offset * (self._n - 1)
        height = self.height - self.offset * (self._n - 1)
        position = self.position + (
            self.width / 2 - width / 2 - (self._n - 1) * self.offset,
            self.height / 2 - height / 2 - (self._n - 1) * self.offset,
        )
        return self._make_subunit_shape(
            position=position, width=width, height=height
        ).position


@dataclasses.dataclass(frozen=True)
class _TextMixin(_SBGNMixin):
    _text: typing.ClassVar[str]
    _font_family: typing.ClassVar[str]
    _font_size_func: typing.ClassVar[typing.Callable]
    _font_style: typing.ClassVar[momapy.drawing.FontStyle] = (
        momapy.drawing.FontStyle.NORMAL
    )
    _font_weight: typing.ClassVar[momapy.drawing.FontWeight | float] = (
        momapy.drawing.FontWeight.NORMAL
    )
    _font_fill: typing.ClassVar[
        momapy.coloring.Color | momapy.drawing.NoneValueType
    ]
    _font_stroke: typing.ClassVar[
        momapy.coloring.Color | momapy.drawing.NoneValueType
    ]

    def _make_text_layout(self):
        text_layout = momapy.core.TextLayout(
            text=self._text,
            position=self.label_center(),
            font_family=self._font_family,
            font_size=self._font_size_func(),
            font_style=self._font_style,
            font_weight=self._font_weight,
            fill=self._font_fill,
            stroke=self._font_stroke,
        )
        return text_layout

    @classmethod
    def _mixin_drawing_elements(cls, obj):
        return obj._make_text_layout().drawing_elements()



"""Tests for momapy.rendering.core module."""
import momapy.rendering.core


def test_renderers_dict_exists():
    """Test that renderers dictionary exists."""
    assert isinstance(momapy.rendering.core.renderers, dict)


def test_register_renderer():
    """Test register_renderer function."""
    class DummyRenderer:
        pass

    momapy.rendering.core.register_renderer("test_renderer", DummyRenderer)
    assert "test_renderer" in momapy.rendering.core.renderers
    assert momapy.rendering.core.renderers["test_renderer"] == DummyRenderer


def test_render_layout_element_exists():
    """Test that render_layout_element function exists."""
    assert hasattr(momapy.rendering.core, 'render_layout_element')
    assert callable(momapy.rendering.core.render_layout_element)


def test_render_layout_elements_exists():
    """Test that render_layout_elements function exists."""
    assert hasattr(momapy.rendering.core, 'render_layout_elements')
    assert callable(momapy.rendering.core.render_layout_elements)



"""Base classes and functions for reading and writing maps"""

import os
import dataclasses
import abc
import typing

import frozendict

readers = {}
writers = {}


def register_reader(name: str, reader_cls: typing.Type):
    """Register a reader"""
    readers[name] = reader_cls


def register_writer(name, writer_cls):
    """Register a writer"""
    writers[name] = writer_cls


@dataclasses.dataclass
class IOResult:
    """Base class for I/O results"""

    exceptions: list[Exception] = dataclasses.field(default_factory=list)


@dataclasses.dataclass
class ReaderResult(IOResult):
    """Base class for reader results"""

    obj: typing.Any | None = None
    annotations: frozendict.frozendict | None = None
    notes: frozendict.frozendict | None = None
    ids: dict | None = None
    file_path: str | os.PathLike | None = None


@dataclasses.dataclass
class WriterResult(IOResult):
    """Base class for writer results"""

    obj: typing.Any | None = None
    file_path: str | os.PathLike | None = None


def read(
    file_path: str | os.PathLike,
    reader: str | None = None,
    **options,
) -> ReaderResult:
    """Read a map file and return a reader result using the given registered reader. If no reader is given, will check for an appropriate reader among the registered readers, using the `check_file` method of each reader. If there is more than one appropriate reader, will use the first one.

    Args:
        file_path: The path of the file to read
        reader: The registered reader
        options: Options to be passed to the reader

    Returns:
        A reader result

    """
    import momapy.io

    momapy.io._ensure_registered()

    reader_cls = None
    if reader is not None:
        reader_cls = readers.get(reader)
        if reader_cls is None:
            raise ValueError(f"no registered reader named '{reader}'")
    else:
        for candidate_reader_cls in readers.values():
            if candidate_reader_cls.check_file(file_path):
                reader_cls = candidate_reader_cls
                break
    if reader_cls is not None:
        result = reader_cls.read(file_path, **options)
    else:
        raise ValueError(
            f"could not find a suitable registered reader for file '{file_path}'"
        )
    return result


def write(
    obj: typing.Any,
    file_path: str | os.PathLike,
    writer: str,
    **options,
) -> WriterResult:
    """Write an object to a file and return a writer result using the given registered writer

    Args:
        obj: The object to write
        file_path: The path of the file to write to
        writer: The registered writer
        options: Options to be passed to the writer

    Returns:
        A writer result
    """
    import momapy.io

    momapy.io._ensure_registered()

    writer_cls = None
    writer_cls = writers.get(writer)
    if writer_cls is None:
        raise ValueError(f"no registered writer named '{writer}'")
    result = writer_cls.write(obj, file_path, **options)
    return result


class Reader(abc.ABC):
    """Base class for readers"""

    @classmethod
    @abc.abstractmethod
    def read(cls, file_path: str | os.PathLike, **options) -> ReaderResult:
        """Read a file and return a reader result using the reader"""
        pass

    @classmethod
    def check_file(cls, file_path: str | os.PathLike) -> bool:
        """Return `true` if the given file is supported by the reader, `false` otherwise"""
        pass


class Writer(abc.ABC):
    """Base class for writers"""

    @classmethod
    @abc.abstractmethod
    def write(
        cls,
        obj: typing.Any,
        file_path: str | os.PathLike,
        **options,
    ) -> WriterResult:
        """Write an object to a file and return a writer result using the writer"""
        pass



import momapy.core
import momapy.builder
import momapy.meta.nodes
import momapy.rendering.core


def render_layout_elements_on_grid(
    output_file,
    layout_elements,
    n_cols,
    width,
    height,
    x_margin=10.0,
    y_margin=10.0,
    format="pdf",
    renderer="skia",
):
    max_x = x_margin * 2 + n_cols * width
    n_rows = len(layout_elements) // n_cols + int(
        bool(len(layout_elements) % n_cols)
    )
    max_y = y_margin * 2 + n_rows * height
    renderer = momapy.rendering.core.renderers[renderer].from_file(
        output_file, max_x, max_y, format
    )
    renderer.begin_session()
    for i, layout_element in enumerate(layout_elements):
        n_row = i // n_cols
        n_col = i % n_cols
        new_position = momapy.geometry.Point(
            x_margin + width * (n_col + 1 / 2),
            y_margin + height * (n_row + 1 / 2),
        )
        if momapy.builder.isinstance_or_builder(
            layout_element, momapy.core.Arc
        ):
            old_position = layout_element.points()[0] * (
                1 / 2
            ) + layout_element.points()[-1] * (1 / 2)
        elif momapy.builder.isinstance_or_builder(
            layout_element, momapy.core.Node
        ):
            old_position = layout_element.position
        else:
            raise TypeError(
                f"layout element must be of type {momapy.core.Node} or {momapy.core.Arc}"
            )
        if not isinstance(layout_element, momapy.builder.Builder):
            layout_element = momapy.builder.builder_from_object(layout_element)
        tx = new_position.x - old_position.x
        ty = new_position.y - old_position.y
        translation = momapy.geometry.Translation(tx, ty)
        if layout_element.transform is None:
            layout_element.transform = momapy.core.TupleBuilder()
        layout_element.transform.append(translation)
        layout_element = momapy.builder.object_from_builder(layout_element)
        renderer.render_layout_element(layout_element)
    renderer.end_session()


def render_nodes_testing(
    output_file,
    node_configs,
    width,
    height,
    x_margin=10.0,
    y_margin=10.0,
    format="pdf",
    renderer="skia",
):
    node_objs = []
    for node_config in node_configs:
        node_cls = node_config[0]
        node_obj = momapy.meta.nodes.Rectangle(
            position=momapy.geometry.Point(0, 0),
            width=10.0,
            height=10.0,
            label=momapy.core.TextLayout(
                position=momapy.geometry.Point(0, 0),
                text=node_cls.__name__,
                font_size=7,
                font_family="Cantarell",
                font_style=momapy.drawing.FontStyle.ITALIC,
                font_weight=momapy.drawing.FontWeight.BOLD,
                fill=momapy.coloring.black,
                stroke=momapy.drawing.NoneValue,
            ),
            border_stroke_width=0.0,
            border_fill=momapy.coloring.white,
        )
        node_objs.append(node_obj)

        kwargs = node_config[1]
        kwargs["position"] = momapy.geometry.Point(0, 0)
        kwargs["fill"] = momapy.coloring.lightblue
        kwargs["stroke"] = momapy.coloring.black
        node_obj = node_cls(**kwargs)
        node_objs.append(node_obj)

        anchor_cross_points = []
        for anchor_name in [
            "north_west",
            "north",
            "north_east",
            "east",
            "south_east",
            "south",
            "south_west",
            "west",
            "center",
            "label_center",
        ]:
            p = getattr(node_obj, anchor_name)()
            cross_point = momapy.meta.nodes.CrossPoint(
                position=p,
                width=5.0,
                height=5.0,
                stroke=momapy.coloring.red,
                stroke_width=0.5,
            )
            anchor_cross_points.append(cross_point)
        kwargs["layout_elements"] = tuple(anchor_cross_points)
        node_obj = node_cls(**kwargs)
        node_objs.append(node_obj)

        angle_cross_points = []
        for angle in range(0, 380, 30):
            p = node_obj.self_angle(angle)
            cross_point = momapy.meta.nodes.CrossPoint(
                position=p,
                width=5.0,
                height=5.0,
                stroke=momapy.coloring.darkgoldenrod,
                stroke_width=0.5,
            )
            angle_cross_points.append(cross_point)
        kwargs["layout_elements"] = tuple(angle_cross_points)
        node_obj = node_cls(**kwargs)
        node_objs.append(node_obj)

    render_layout_elements_on_grid(
        output_file=output_file,
        layout_elements=node_objs,
        n_cols=4,
        width=width,
        height=height,
        x_margin=x_margin,
        y_margin=y_margin,
        format=format,
        renderer=renderer,
    )


def render_arcs_testing(
    output_file,
    arc_configs,
    width,
    height,
    x_margin=10.0,
    y_margin=10.0,
    format="pdf",
    renderer="skia",
    path_stroke=None,
    path_stroke_width=None,
    path_fill=None,
    arrowhead_stroke=None,
    arrowhead_stroke_width=None,
):
    arc_objs = []
    for arc_config in arc_configs:
        arc_cls = arc_config[0]
        arc_obj = momapy.meta.nodes.Rectangle(
            position=momapy.geometry.Point(0, 0),
            width=width,
            height=height,
            stroke_width=0.0,
            fill=momapy.drawing.NoneValue,
            label=momapy.core.TextLayout(
                position=momapy.geometry.Point(0, 0),
                text=arc_cls.__name__,
                font_size=10,
                font_family="Cantarell",
                fill=momapy.coloring.black,
                stroke=momapy.drawing.NoneValue,
            ),
        )
        arc_objs.append(arc_obj)

        kwargs = arc_config[1]
        if "path_stroke" not in kwargs:
            kwargs["path_stroke"] = path_stroke
        if "path_stroke_width" not in kwargs:
            kwargs["path_stroke_width"] = path_stroke_width
        if "path_fill" not in kwargs:
            kwargs["path_fill"] = path_fill

        if "arrowhead_stroke" not in kwargs:
            kwargs["arrowhead_stroke"] = arrowhead_stroke
        if "arrowhead_stroke_width" not in kwargs:
            kwargs["arrowhead_stroke_width"] = arrowhead_stroke_width
        kwargs["segments"] = tuple(
            [
                momapy.geometry.Segment(
                    momapy.geometry.Point(0, 0),
                    momapy.geometry.Point(0.8 * width, -0.8 * height),
                )
            ]
        )
        arc_obj = arc_cls(**kwargs)
        arc_objs.append(arc_obj)

        anchor_cross_points = []
        for anchor_name in [
            "start_point",
            "end_point",
            "arrowhead_base",
            "arrowhead_tip",
        ]:
            p = getattr(arc_obj, anchor_name)()
            cross_point = momapy.meta.nodes.CrossPoint(
                position=p,
                width=5.0,
                height=5.0,
                stroke=momapy.coloring.red,
                stroke_width=0.5,
            )
            anchor_cross_points.append(cross_point)
        kwargs["layout_elements"] = tuple(anchor_cross_points)
        arc_obj = arc_cls(**kwargs)
        arc_objs.append(arc_obj)

        fraction_cross_points = []
        for fraction in numpy.arange(0, 1.1, 0.2):
            p, _ = arc_obj.fraction(fraction)
            cross_point = momapy.meta.nodes.CrossPoint(
                position=p,
                width=5.0,
                height=5.0,
                stroke=momapy.coloring.darkgoldenrod,
                stroke_width=0.5,
            )
            fraction_cross_points.append(cross_point)
        kwargs["layout_elements"] = tuple(fraction_cross_points)
        arc_obj = arc_cls(**kwargs)
        arc_objs.append(arc_obj)

    render_layout_elements_on_grid(
        output_file=output_file,
        layout_elements=arc_objs,
        n_cols=4,
        width=width,
        height=height,
        x_margin=x_margin,
        y_margin=y_margin,
        format=format,
        renderer=renderer,
    )



"""Class for rendering with Cairo"""

import dataclasses
import typing
import typing_extensions
import math
import os

try:
    import cairo
    import gi

    gi.require_version("Pango", "1.0")
    gi.require_version("PangoCairo", "1.0")
    import gi.repository.Pango
    import gi.repository.PangoCairo
except ModuleNotFoundError as e:
    raise Exception(
        "You might want to install momapy with the cairo extra: momapy[cairo]"
    ) from e


import momapy.drawing
import momapy.geometry
import momapy.rendering.core


@dataclasses.dataclass(kw_only=True)
class CairoRenderer(momapy.rendering.core.StatefulRenderer):
    """Class for cairo renderers"""

    formats: typing.ClassVar[list[str]] = ["pdf", "svg", "png", "ps"]
    _de_class_func_mapping: typing.ClassVar[dict] = {
        momapy.drawing.Group: "_render_group",
        momapy.drawing.Path: "_render_path",
        momapy.drawing.Text: "_render_text",
        momapy.drawing.Ellipse: "_render_ellipse",
        momapy.drawing.Rectangle: "_render_rectangle",
    }
    _pa_class_func_mapping: typing.ClassVar[dict] = {
        momapy.drawing.MoveTo: "_add_move_to",
        momapy.drawing.LineTo: "_add_line_to",
        momapy.drawing.CurveTo: "_add_curve_to",
        momapy.drawing.QuadraticCurveTo: "_add_quadratic_curve_to",
        momapy.drawing.ClosePath: "_add_close_path",
        momapy.drawing.EllipticalArc: "_add_elliptical_arc",
    }
    _tr_class_func_mapping: typing.ClassVar[dict] = {
        momapy.geometry.Translation: "_add_translation",
        momapy.geometry.Rotation: "_add_rotation",
        momapy.geometry.Scaling: "_add_scaling",
        momapy.geometry.MatrixTransformation: "_add_matrix_transformation",
    }
    _te_font_style_slant_mapping: typing.ClassVar[dict] = {
        momapy.drawing.FontStyle.NORMAL: gi.repository.Pango.Style.NORMAL,
        momapy.drawing.FontStyle.ITALIC: gi.repository.Pango.Style.ITALIC,
        momapy.drawing.FontStyle.OBLIQUE: gi.repository.Pango.Style.OBLIQUE,
    }
    context: cairo.Context = dataclasses.field(
        metadata={"description": "A cairo context"}
    )
    _config: dict = dataclasses.field(default_factory=dict)
    _pango_font_descriptions: dict = dataclasses.field(default_factory=dict)

    @classmethod
    def from_file(
        cls,
        file_path: str | os.PathLike,
        width: float,
        height: float,
        format_: typing.Literal["pdf", "svg", "png", "ps"] = "pdf",
    ) -> typing_extensions.Self:
        config = {}
        if format_ == "pdf":
            surface = cairo.PDFSurface(file_path, width, height)
        elif format_ == "ps":
            surface = cairo.PSSurface(file_path, width, height)
        elif format_ == "svg":
            surface = cairo.SVGSurface(file_path, width, height)
        elif format_ == "png":
            surface = cairo.ImageSurface(cairo.FORMAT_ARGB32, int(width), int(height))
        config["surface"] = surface
        config["file_path"] = file_path
        config["width"] = width
        config["height"] = height
        config["format"] = format_
        context = cairo.Context(surface)
        return cls(context=context, _config=config)

    def begin_session(self):
        pass

    def end_session(self):
        surface = self.context.get_target()
        format_ = self._config.get("format")
        if format_ == "png":
            surface.write_to_png(self._config["file_path"])
        surface.finish()
        surface.flush()

    def new_page(self, width, height):
        format_ = self._config.get("format")
        if format_ == "pdf" or format_ == "ps":
            self.context.show_page()
            surface = self.context.get_target()
            surface.set_size(width, height)

    def render_map(self, map_):
        self.render_layout_element(map_.layout)

    def render_layout_element(self, layout_element):
        drawing_elements = layout_element.drawing_elements()
        for drawing_element in drawing_elements:
            self.render_drawing_element(drawing_element)

    def render_drawing_element(self, drawing_element):
        self.save()
        self.set_current_state_from_drawing_element(drawing_element)
        self._add_transform_from_drawing_element(drawing_element)
        class_ = type(drawing_element)
        if issubclass(class_, momapy.builder.Builder):
            class_ = class_._cls_to_build
        de_func = getattr(self, self._de_class_func_mapping[class_])
        de_func(drawing_element)
        self.restore()

    def self_save(self):
        self.context.save()

    def self_restore(self):
        self.context.restore()
        self.context.new_path()  # context.restore() does not forget the current path

    def _make_stroke_paint(self):
        """Set stroke paint on context"""
        stroke = self.get_current_value("stroke")
        stroke_width = self.get_current_value("stroke_width")
        stroke_dasharray = self.get_current_value("stroke_dasharray")
        stroke_dashoffset = self.get_current_value("stroke_dashoffset")

        if stroke != momapy.drawing.NoneValue and stroke is not None:
            self.context.set_line_width(stroke_width)
            self.context.set_source_rgba(*stroke.to_rgba(rgba_range=(0, 1)))
            if (
                stroke_dasharray is not None
                and stroke_dasharray != momapy.drawing.NoneValue
            ):
                self.context.set_dash(stroke_dasharray, stroke_dashoffset or 0)
            else:
                self.context.set_dash([])
            return True
        return False

    def _make_fill_paint(self):
        """Set fill paint on context"""
        fill = self.get_current_value("fill")
        if fill != momapy.drawing.NoneValue and fill is not None:
            self.context.set_source_rgba(*fill.to_rgba(rgba_range=(0, 1)))
            return True
        return False

    def _stroke_and_fill(self):
        """Apply stroke and fill to current path"""
        has_fill = self._make_fill_paint()
        has_stroke = self.get_current_value("stroke") not in [
            momapy.drawing.NoneValue,
            None,
        ]

        if has_fill:
            if has_stroke:
                self.context.fill_preserve()
            else:
                self.context.fill()

        if has_stroke:
            self._make_stroke_paint()
            self.context.stroke()

    def _add_transform_from_drawing_element(self, drawing_element):
        if (
            drawing_element.transform is not None
            and drawing_element.transform != momapy.drawing.NoneValue
        ):
            for transformation in drawing_element.transform:
                self._add_transformation(transformation)

    def _add_transformation(self, transformation):
        class_ = type(transformation)
        if issubclass(class_, momapy.builder.Builder):
            class_ = class_._cls_to_build
        tr_func = getattr(self, self._tr_class_func_mapping[class_])
        return tr_func(transformation)

    def _render_group(self, group):
        for drawing_element in group.elements:
            self.render_drawing_element(drawing_element)

    def _add_path_action_to_context(self, path_action):
        class_ = type(path_action)
        if issubclass(class_, momapy.builder.Builder):
            class_ = class_._cls_to_build
        pa_func = getattr(self, self._pa_class_func_mapping[class_])
        pa_func(path_action)

    def _render_path(self, path):
        for action in path.actions:
            self._add_path_action_to_context(action)
        self._stroke_and_fill()

    def _render_text(self, text):
        pango_layout = gi.repository.PangoCairo.create_layout(self.context)

        font_family = self.get_current_value("font_family")
        font_size = self.get_current_value("font_size")
        font_weight = self.get_current_value("font_weight")
        font_style = self.get_current_value("font_style")

        pango_font_description = self._pango_font_descriptions.get(
            (font_family, font_size, font_weight, font_style)
        )

        if pango_font_description is None:
            pango_font_description = gi.repository.Pango.FontDescription()
            pango_font_description.set_family(font_family)
            pango_font_description.set_absolute_size(
                gi.repository.Pango.units_from_double(font_size)
            )
            pango_font_description.set_weight(int(font_weight))
            pango_style = self._te_font_style_slant_mapping.get(
                font_style, gi.repository.Pango.Style.NORMAL
            )
            pango_font_description.set_style(pango_style)
            self._pango_font_descriptions[
                (font_family, font_size, font_weight, font_style)
            ] = pango_font_description

        pango_layout.set_font_description(pango_font_description)
        pango_layout.set_text(text.text)
        pos = pango_layout.index_to_pos(0)
        x_offset = gi.repository.Pango.units_to_double(pos.x)
        pango_layout_iter = pango_layout.get_iter()
        baseline = pango_layout_iter.get_baseline()
        y_offset = gi.repository.Pango.units_to_double(baseline)
        self.context.translate(text.x - x_offset, text.y - y_offset)
        if self._make_fill_paint():
            gi.repository.PangoCairo.show_layout(self.context, pango_layout)
        if self._make_stroke_paint():
            gi.repository.PangoCairo.layout_path(self.context, pango_layout)
            self.context.stroke()

    def _render_ellipse(self, ellipse):
        self.context.save()
        self.context.translate(ellipse.x, ellipse.y)
        self.context.scale(ellipse.rx, ellipse.ry)
        self.context.arc(0, 0, 1, 0, 2 * math.pi)
        self.context.close_path()
        self.context.restore()
        self._stroke_and_fill()

    def _render_rectangle(self, rectangle):
        path = rectangle.to_path()
        self._render_path(path)

    def _add_move_to(self, move_to):
        self.context.move_to(move_to.x, move_to.y)

    def _add_line_to(self, line_to):
        self.context.line_to(line_to.x, line_to.y)

    def _add_curve_to(self, curve_to):
        self.context.curve_to(
            curve_to.control_point1.x,
            curve_to.control_point1.y,
            curve_to.control_point2.x,
            curve_to.control_point2.y,
            curve_to.x,
            curve_to.y,
        )

    def _add_quadratic_curve_to(self, quadratic_curve_to):
        cairo_current_point = self.context.get_current_point()
        current_point = momapy.geometry.Point(
            cairo_current_point[0], cairo_current_point[1]
        )
        curve_to = quadratic_curve_to.to_curve_to(current_point)
        self._add_curve_to(curve_to)

    def _add_close_path(self, close_path):
        self.context.close_path()

    def _add_elliptical_arc(self, elliptical_arc):
        # Get current point
        current_point = self.context.get_current_point()
        p1 = momapy.geometry.Point(current_point[0], current_point[1])

        # Create elliptical arc geometry object
        obj = momapy.geometry.EllipticalArc(
            p1=p1,
            p2=elliptical_arc.point,
            rx=elliptical_arc.rx,
            ry=elliptical_arc.ry,
            x_axis_rotation=elliptical_arc.x_axis_rotation,
            arc_flag=elliptical_arc.arc_flag,
            sweep_flag=elliptical_arc.sweep_flag,
        )

        # Get center parameterization
        cx, cy, rx, ry, sigma, theta1, theta2, delta_theta = (
            obj.get_center_parameterization()
        )

        # Apply transformation for rotated ellipse
        self.context.save()
        self.context.translate(cx, cy)
        self.context.rotate(sigma)
        self.context.scale(rx, ry)

        # Draw arc
        if delta_theta > 0:
            self.context.arc(0, 0, 1, theta1, theta2)
        else:
            self.context.arc_negative(0, 0, 1, theta1, theta2)

        self.context.restore()

    def _add_translation(self, translation):
        self.context.translate(translation.tx, translation.ty)

    def _add_rotation(self, rotation):
        if rotation.point is not None:
            self.context.translate(rotation.point.x, rotation.point.y)
            self.context.rotate(rotation.angle)
            self.context.translate(-rotation.point.x, -rotation.point.y)
        else:
            self.context.rotate(rotation.angle)

    def _add_scaling(self, scaling):
        self.context.scale(scaling.sx, scaling.sy)

    def _add_matrix_transformation(self, matrix_transformation):
        m = cairo.Matrix(
            xx=matrix_transformation.m[0][0],
            yx=matrix_transformation.m[1][0],
            xy=matrix_transformation.m[0][1],
            yy=matrix_transformation.m[1][1],
            x0=matrix_transformation.m[0][2],
            y0=matrix_transformation.m[1][2],
        )
        self.context.transform(m)


momapy.rendering.core.register_renderer("cairo", CairoRenderer)
